"""
Data sources for SpikeInterface
"""

from .sourcebase import BaseDataSource
import sys
import logging

import numpy as np

try:
    from distutils.version import LooseVersion as V
    import spikeinterface
    if V(spikeinterface.__version__)>='0.90.1':
        HAVE_SI = True
    else:
        HAVE_SI = False
except ImportError:
    HAVE_SI = False


from .signals import BaseAnalogSignalSource
from .spikes import BaseSpikeSource


class SpikeInterfaceRecordingSource(BaseAnalogSignalSource):
    def __init__(self, recording, segment_index=0):
        BaseAnalogSignalSource.__init__(self)

        self.recording = recording
        self.segment_index = segment_index

        self._nb_channel = self.recording.get_num_channels()
        self.sample_rate = self.recording.get_sampling_frequency()

    @property
    def nb_channel(self):
        return self._nb_channel

    def get_channel_name(self, chan=0):
        return str(self.recording.channel_ids[chan])

    @property
    def t_start(self):
        return 0.

    @property
    def t_stop(self):
        return self.get_length() / self.sample_rate

    def get_length(self):
        return self.recording.get_num_samples(segment_index=self.segment_index)

    def get_shape(self):
        return (self.get_length(),self.nb_channel)

    def get_chunk(self, i_start=None, i_stop=None):
        traces = self.recording.get_traces(segment_index=self.segment_index, start_frame=i_start, end_frame=i_stop)
        return traces

    def time_to_index(self, t):
        return int(t * self.sample_rate)

    def index_to_time(self, ind):
        return float(ind / self.sample_rate)



class SpikeInterfaceSortingSource(BaseSpikeSource):
    def __init__(self, sorting, segment_index=0):
        BaseSpikeSource.__init__(self)

        self.sorting = sorting
        self.segment_index = segment_index

        #TODO
        self._t_stop = 10.

    @property
    def nb_channel(self):
        return len(self.sorting.unit_ids)

    def get_channel_name(self, chan=0):
        return str(self.sorting.unit_ids[chan])

    @property
    def t_start(self):
        return 0.

    @property
    def t_stop(self):
        return self._t_stop

    def get_chunk(self, chan=0,  i_start=None, i_stop=None):
        unit_id = self.sorting.unit_ids[chan]
        spike_frames = self.sorting.get_unit_spike_train(unit_id,
                    segment_index=self.segment_index, start_frame=i_start, end_frame=i_stop)
        spike_frames = spike_frames[i_start:i_stop]
        spike_times = spike_frames / self.sorting.get_sampling_frequency()
        return spike_times

    def get_chunk_by_time(self, chan=0,  t_start=None, t_stop=None):
        spike_times = self.get_chunk(chan=chan)
        i1 = np.searchsorted(spike_times, t_start, side='left')
        i2 = np.searchsorted(spike_times, t_stop, side='left')
        sl = slice(i1, i2+1)
        return spike_times[sl]






from ephyviewer.standalone import WindowManager, RawNeoOpenDialog
import pyqtgraph as pg



def test_WindowManager(interactive=False):
    app = pg.mkQApp()
    manager = WindowManager()

    if interactive:
        manager.open_dialog()
        if manager.windows:
            app.exec_()


def test_RawNeoOpenDialog(interactive=False):
    app = pg.mkQApp()
    dia = RawNeoOpenDialog()

    if interactive:
        dia.show()
        app.exec_()

    print(dia.final_params)



if __name__=='__main__':
    test_WindowManager(interactive=True)
    test_RawNeoOpenDialog(interactive=True)



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)
import os
import numpy as np

import matplotlib.cm
import matplotlib.colors

try:
    import pandas as pd
    HAVE_PANDAS = True
except ImportError:
    HAVE_PANDAS = False


from .sourcebase import BaseDataSource
from .events import BaseEventAndEpoch



class InMemoryEpochSource(BaseEventAndEpoch):
    type = 'Epoch'

    def __init__(self, all_epochs=[]):
        BaseEventAndEpoch.__init__(self, all=all_epochs)

        s = [ np.max(e['time']+e['duration']) for e in self.all if len(e['time'])>0]
        self._t_stop = max(s) if len(s)>0 else 0

    def get_chunk(self, chan=0,  i_start=None, i_stop=None):
        ep_times = self.all[chan]['time'][i_start:i_stop]
        ep_durations = self.all[chan]['duration'][i_start:i_stop]
        ep_labels = self.all[chan]['label'][i_start:i_stop]
        return ep_times, ep_durations, ep_labels

    def get_chunk_by_time(self, chan=0,  t_start=None, t_stop=None):
        ep_times = self.all[chan]['time']
        ep_durations = self.all[chan]['duration']
        ep_labels = self.all[chan]['label']

        keep1 = (ep_times>=t_start) & (ep_times<=t_stop) # epochs that start inside range
        keep2 = (ep_times+ep_durations>=t_start) & (ep_times+ep_durations<=t_stop) # epochs that end inside range
        keep3 = (ep_times<=t_start) & (ep_times+ep_durations>=t_stop) # epochs that span the range
        keep = keep1 | keep2 | keep3

        return ep_times[keep], ep_durations[keep], ep_labels[keep]




class WritableEpochSource(InMemoryEpochSource):
    """
    Identique to EpochSource but onlye one channel that can be persisently saved.

    epoch is dict
    { 'time':np.array, 'duration':np.array, 'label':np.array, 'name':' ''}
    """
    def __init__(self, epoch=None, possible_labels=[], color_labels=None, channel_name='', restrict_to_possible_labels=False):

        self.possible_labels = possible_labels
        self.channel_name = channel_name

        if epoch is None:
            epoch = self.load()

        InMemoryEpochSource.__init__(self, all_epochs=[epoch])

        # assign each epoch a fixed, unique integer id
        self._next_id = 0
        for chan in self.all:
            chan['id'] = np.arange(self._next_id, self._next_id + len(chan['time']))
            self._next_id += len(chan['time'])

        assert self.all[0]['time'].dtype.kind=='f'
        assert self.all[0]['duration'].dtype.kind=='f'

        # add labels missing from possible_labels but found in epoch data
        new_labels_from_data = list(set(epoch['label'])-set(self.possible_labels))
        if restrict_to_possible_labels:
            assert len(new_labels_from_data)==0, f'epoch data contains labels not found in possible_labels: {new_labels_from_data}'
        self.possible_labels += new_labels_from_data

        # put the epochs into a canonical order after loading
        self._clean_and_set(self.all[0]['time'], self.all[0]['duration'], self.all[0]['label'], self.all[0]['id'])

        # TODO: colors should be managed directly by EpochEncoder
        if color_labels is None:
            n = len(self.possible_labels)
            cmap = matplotlib.cm.get_cmap('Dark2' , n)
            color_labels = [ matplotlib.colors.ColorConverter().to_rgb(cmap(i)) for i in  range(n)]
            color_labels = (np.array(color_labels)*255).astype(int)
            color_labels = color_labels.tolist()
        self.color_labels = color_labels

    @property
    def ep_times(self):
        return self.all[0]['time']

    @ep_times.setter
    def ep_times(self, arr):
        self.all[0]['time'] = arr

    @property
    def ep_durations(self):
        return self.all[0]['duration']

    @ep_durations.setter
    def ep_durations(self, arr):
        self.all[0]['duration'] = arr

    @property
    def ep_labels(self):
        return self.all[0]['label']

    @ep_labels.setter
    def ep_labels(self, arr):
        self.all[0]['label'] = arr

    @property
    def ep_ids(self):
        return self.all[0]['id']

    @ep_ids.setter
    def ep_ids(self, arr):
        self.all[0]['id'] = arr

    @property
    def ep_stops(self):
        return self.ep_times + self.ep_durations

    @property
    def id_to_ind(self):
        return dict((id,ind) for ind,id in enumerate(self.ep_ids))

    def get_chunk(self, chan=0,  i_start=None, i_stop=None):
        assert chan==0
        ep_times = self.all[chan]['time'][i_start:i_stop]
        ep_durations = self.all[chan]['duration'][i_start:i_stop]
        ep_labels = self.all[chan]['label'][i_start:i_stop]
        ep_ids = self.all[chan]['id'][i_start:i_stop]
        return ep_times, ep_durations, ep_labels, ep_ids


    def get_chunk_by_time(self, chan=0,  t_start=None, t_stop=None):
        assert chan==0
        ep_times = self.all[chan]['time']
        ep_durations = self.all[chan]['duration']
        ep_labels = self.all[chan]['label']
        ep_ids = self.all[chan]['id']

        keep1 = (ep_times>=t_start) & (ep_times<=t_stop) # epochs that start inside range
        keep2 = (ep_times+ep_durations>=t_start) & (ep_times+ep_durations<=t_stop) # epochs that end inside range
        keep3 = (ep_times<=t_start) & (ep_times+ep_durations>=t_stop) # epochs that span the range
        keep = keep1 | keep2 | keep3

        return ep_times[keep], ep_durations[keep], ep_labels[keep], ep_ids[keep]

    @property
    def label_to_color(self):
        return dict(zip(self.possible_labels, self.color_labels))

    def color_by_label(self, label):
        return self.label_to_color[label]

    def _clean_and_set(self, ep_times, ep_durations, ep_labels, ep_ids):

        # remove bad epochs
        keep = ep_durations >= 1e-6 # discard epochs shorter than 1 microsecond or with negative duration
        ep_times = ep_times[keep]
        ep_durations = ep_durations[keep]
        ep_labels = ep_labels[keep]
        ep_ids = ep_ids[keep]

        # sort epochs by start time
        ordering = np.argsort(ep_times)
        ep_times = ep_times[ordering]
        ep_durations = ep_durations[ordering]
        ep_labels = ep_labels[ordering]
        ep_ids = ep_ids[ordering]

        # set epochs for the WritableEpochSource
        self.ep_times = ep_times
        self.ep_durations = ep_durations
        self.ep_labels = ep_labels
        self.ep_ids = ep_ids

    def add_epoch(self, t1, duration, label):

        ep_times, ep_durations, ep_labels, ep_ids = self.ep_times, self.ep_durations, self.ep_labels, self.ep_ids
        ep_times = np.append(ep_times, t1)
        ep_durations = np.append(ep_durations, duration)
        ep_labels = np.append(ep_labels, label)
        ep_ids = np.append(ep_ids, self._next_id)
        self._next_id += 1

        self._clean_and_set(ep_times, ep_durations, ep_labels, ep_ids)

    def delete_epoch(self, ind):

        ep_times, ep_durations, ep_labels, ep_ids = self.ep_times, self.ep_durations, self.ep_labels, self.ep_ids
        ep_times = np.delete(ep_times, ind)
        ep_durations = np.delete(ep_durations, ind)
        ep_labels = np.delete(ep_labels, ind)
        ep_ids = np.delete(ep_ids, ind)

        self._clean_and_set(ep_times, ep_durations, ep_labels, ep_ids)

    def delete_in_between(self, t1, t2):

        ep_times, ep_durations, ep_stops, ep_labels, ep_ids = self.ep_times, self.ep_durations, self.ep_stops, self.ep_labels, self.ep_ids

        for i in range(len(ep_times)):

            # if epoch starts and ends inside range, delete it
            if ep_times[i]>=t1 and ep_stops[i]<=t2:
                ep_durations[i] = -1 # non-positive duration flags this epoch for clean up

            # if epoch starts before and ends inside range, truncate it
            elif ep_times[i]<t1 and (t1<ep_stops[i]<=t2):
                ep_durations[i] = t1 - ep_times[i]

            # if epoch starts inside and ends after range, truncate it
            elif (t1<=ep_times[i]<t2) and ep_stops[i]>t2:
                ep_durations[i] = ep_stops[i] - t2
                ep_times[i] = t2

            # if epoch starts before and ends after range,
            # truncate the first part and add a new epoch for the end part
            elif ep_times[i]<=t1 and ep_stops[i]>=t2:
                ep_durations[i] = t1 - ep_times[i]
                ep_times = np.append(ep_times, t2)
                ep_durations = np.append(ep_durations, ep_stops[i]-t2)
                ep_labels = np.append(ep_labels, ep_labels[i])
                ep_ids = np.append(ep_ids, self._next_id)
                self._next_id += 1

        self._clean_and_set(ep_times, ep_durations, ep_labels, ep_ids)

    def merge_neighbors(self):

        ep_times, ep_durations, ep_stops, ep_labels, ep_ids = self.ep_times, self.ep_durations, self.ep_stops, self.ep_labels, self.ep_ids

        for label in self.possible_labels:
            inds, = np.nonzero(ep_labels == label)
            for i in range(len(inds)-1):

                # if two sequentially adjacent epochs with the same label
                # overlap or have less than 1 microsecond separation, merge them
                if ep_times[inds[i+1]] - ep_stops[inds[i]] < 1e-6:

                    # stretch the second epoch to cover the range of both epochs
                    ep_times[inds[i+1]] = min(ep_times[inds[i]], ep_times[inds[i+1]])
                    ep_stops[inds[i+1]] = max(ep_stops[inds[i]], ep_stops[inds[i+1]])
                    ep_durations[inds[i+1]] = ep_stops[inds[i+1]] - ep_times[inds[i+1]]

                    # delete the first epoch
                    ep_durations[inds[i]] = -1 # non-positive duration flags this epoch for clean up

        self._clean_and_set(ep_times, ep_durations, ep_labels, ep_ids)

    def split_epoch(self, ind, t_split):

        ep_times, ep_durations, ep_stops, ep_labels, ep_ids = self.ep_times, self.ep_durations, self.ep_stops, self.ep_labels, self.ep_ids

        if t_split <= ep_times[ind] or ep_stops[ind] <= t_split:
            return

        ep_durations[ind] = t_split - ep_times[ind]
        ep_times = np.append(ep_times, t_split)
        ep_durations = np.append(ep_durations, ep_stops[ind]-t_split)
        ep_labels = np.append(ep_labels, ep_labels[ind])
        ep_ids = np.append(ep_ids, self._next_id)
        self._next_id += 1

        self._clean_and_set(ep_times, ep_durations, ep_labels, ep_ids)

    def fill_blank(self, method='from_left'):

        ep_times, ep_durations, ep_labels, ep_ids = self.ep_times, self.ep_durations, self.ep_labels, self.ep_ids

        mask = ((ep_times[:-1] + ep_durations[:-1])<ep_times[1:])
        inds,  = np.nonzero(mask)

        if method=='from_left':
            for ind in inds:
                ep_durations[ind] = ep_times[ind+1] - ep_times[ind]

        elif method=='from_right':
            for ind in inds:
                gap = ep_times[ind+1] - (ep_times[ind] + ep_durations[ind])
                ep_times[ind+1] -= gap
                ep_durations[ind+1] += gap

        elif method=='from_nearest':
            for ind in inds:
                gap = ep_times[ind+1] - (ep_times[ind] + ep_durations[ind])
                ep_durations[ind] += gap/2.
                ep_times[ind+1] -= gap/2.
                ep_durations[ind+1] += gap/2.


        self._clean_and_set(ep_times, ep_durations, ep_labels, ep_ids)

    def load(self):
        """
        Returns a dictionary containing the data for an epoch.

        Derived subclasses of WritableEpochSource override this method to
        implement loading a file or importing data from objects in memory. The
        superclass implementation WritableEpochSource.load() creates an empty
        dictionary with the correct keys and types. It can be called from the
        subclass implementation using super().load() if, for example, the file
        to be loaded does not exist.

        The method returns a dictionary containing the loaded data in this form:

        { 'time': np.array, 'duration': np.array, 'label': np.array, 'name': string }
        """

        s = max([len(l) for l in self.possible_labels])
        epoch = {'time':     np.array([], dtype='float64'),
                 'duration': np.array([], dtype='float64'),
                 'label':    np.array([], dtype='U'+str(s)),
                 'name':     self.channel_name}
        return epoch

    def save(self):
        print('WritableEpochSource.save')
        raise NotImplementedError()



class CsvEpochSource(WritableEpochSource):
    def __init__(self, filename, possible_labels, color_labels=None, channel_name='', restrict_to_possible_labels=False):
        assert HAVE_PANDAS, 'Pandas is not installed'

        self.filename = filename

        WritableEpochSource.__init__(self, epoch=None, possible_labels=possible_labels, color_labels=color_labels, channel_name=channel_name, restrict_to_possible_labels=restrict_to_possible_labels)

    def load(self):
        """
        Returns a dictionary containing the data for an epoch.

        Data is loaded from the CSV file if it exists; otherwise the superclass
        implementation in WritableEpochSource.load() is called to create an
        empty dictionary with the correct keys and types.

        The method returns a dictionary containing the loaded data in this form:

        { 'time': np.array, 'duration': np.array, 'label': np.array, 'name': string }
        """

        if os.path.exists(self.filename):
            # if file already exists, load previous epoch
            df = pd.read_csv(self.filename,  index_col=None, dtype={
                'time':     'float64',
                'duration': 'float64',
                'label':    'U'})
            epoch = {'time':     df['time'].values,
                     'duration': df['duration'].values,
                     'label':    df['label'].values,
                     'name':     self.channel_name}
        else:
            # if file does NOT already exist, use superclass method for creating
            # an empty dictionary
            epoch = super().load()

        return epoch

    def save(self):
        df = pd.DataFrame()
        df['time'] = np.round(self.ep_times, 6)         # round to nearest microsecond
        df['duration'] = np.round(self.ep_durations, 6) # round to nearest microsecond
        df['label'] = self.ep_labels
        df.sort_values(['time', 'duration', 'label'], inplace=True)
        df.to_csv(self.filename, index=False)



# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 5.15.2
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore

qt_resource_data = b"\
\x00\x00\x09q\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0d\x0a<svg\x0d\x0a   \
xmlns:dc=\x22http:/\
/purl.org/dc/ele\
ments/1.1/\x22\x0d\x0a   \
xmlns:cc=\x22http:/\
/creativecommons\
.org/ns#\x22\x0d\x0a   xm\
lns:rdf=\x22http://\
www.w3.org/1999/\
02/22-rdf-syntax\
-ns#\x22\x0d\x0a   xmlns:\
svg=\x22http://www.\
w3.org/2000/svg\x22\
\x0d\x0a   xmlns=\x22http\
://www.w3.org/20\
00/svg\x22\x0d\x0a   xmln\
s:sodipodi=\x22http\
://sodipodi.sour\
ceforge.net/DTD/\
sodipodi-0.dtd\x22\x0d\
\x0a   xmlns:inksca\
pe=\x22http://www.i\
nkscape.org/name\
spaces/inkscape\x22\
\x0d\x0a   inkscape:ve\
rsion=\x221.0 (4035\
a4fb49, 2020-05-\
01)\x22\x0d\x0a   sodipod\
i:docname=\x22epoch\
-encoder-seek.sv\
g\x22\x0d\x0a   id=\x22svg6\x22\
\x0d\x0a   version=\x221.\
1\x22\x0d\x0a   viewBox=\x22\
0 0 16 16\x22>\x0d\x0a  <\
metadata\x0d\x0a     i\
d=\x22metadata10\x22>\x0d\
\x0a    <rdf:RDF>\x0d\x0a\
      <cc:Work\x0d\x0a\
         rdf:abo\
ut=\x22\x22>\x0d\x0a        \
<dc:format>image\
/svg+xml</dc:for\
mat>\x0d\x0a        <d\
c:type\x0d\x0a        \
   rdf:resource=\
\x22http://purl.org\
/dc/dcmitype/Sti\
llImage\x22 />\x0d\x0a   \
   </cc:Work>\x0d\x0a \
   </rdf:RDF>\x0d\x0a \
 </metadata>\x0d\x0a  \
<sodipodi:namedv\
iew\x0d\x0a     inksca\
pe:current-layer\
=\x22g40\x22\x0d\x0a     ink\
scape:window-max\
imized=\x221\x22\x0d\x0a    \
 inkscape:window\
-y=\x2232\x22\x0d\x0a     in\
kscape:window-x=\
\x22-8\x22\x0d\x0a     inksc\
ape:cy=\x227.146120\
2\x22\x0d\x0a     inkscap\
e:cx=\x221.3121066\x22\
\x0d\x0a     inkscape:\
zoom=\x2232\x22\x0d\x0a     \
inkscape:bbox-pa\
ths=\x22false\x22\x0d\x0a   \
  inkscape:bbox-\
nodes=\x22true\x22\x0d\x0a  \
   inkscape:snap\
-bbox-edge-midpo\
ints=\x22true\x22\x0d\x0a   \
  inkscape:snap-\
bbox-midpoints=\x22\
true\x22\x0d\x0a     inks\
cape:snap-bbox=\x22\
true\x22\x0d\x0a     show\
grid=\x22false\x22\x0d\x0a  \
   id=\x22namedview\
8\x22\x0d\x0a     inkscap\
e:window-height=\
\x221017\x22\x0d\x0a     ink\
scape:window-wid\
th=\x221920\x22\x0d\x0a     \
inkscape:pagesha\
dow=\x222\x22\x0d\x0a     in\
kscape:pageopaci\
ty=\x220\x22\x0d\x0a     gui\
detolerance=\x2210\x22\
\x0d\x0a     gridtoler\
ance=\x2210\x22\x0d\x0a     \
objecttolerance=\
\x2210\x22\x0d\x0a     borde\
ropacity=\x221\x22\x0d\x0a  \
   bordercolor=\x22\
#666666\x22\x0d\x0a     p\
agecolor=\x22#fffff\
f\x22\x0d\x0a     inkscap\
e:document-rotat\
ion=\x220\x22 />\x0d\x0a  <d\
efs\x0d\x0a     id=\x22de\
fs3051\x22>\x0d\x0a    <s\
tyle\x0d\x0a       id=\
\x22current-color-s\
cheme\x22\x0d\x0a       t\
ype=\x22text/css\x22>\x0d\
\x0a      .ColorSch\
eme-Text {\x0d\x0a    \
    color:#23262\
9;\x0d\x0a      }\x0d\x0a   \
   </style>\x0d\x0a  <\
/defs>\x0d\x0a  <g\x0d\x0a  \
   transform=\x22tr\
anslate(0.031249\
78,-0.15625)\x22\x0d\x0a \
    id=\x22g40\x22>\x0d\x0a \
   <rect\x0d\x0a      \
 style=\x22color:#0\
00000;overflow:v\
isible;fill:#232\
629;stroke-width\
:23.968;stroke-l\
inecap:round\x22\x0d\x0a \
      id=\x22rect12\
\x22\x0d\x0a       width=\
\x224.9142199\x22\x0d\x0a   \
    height=\x229.90\
02571\x22\x0d\x0a       x\
=\x225.6991401\x22\x0d\x0a  \
     y=\x22-9.91887\
86\x22\x0d\x0a       tran\
sform=\x22rotate(90\
)\x22 />\x0d\x0a    <path\
\x0d\x0a       id=\x22pat\
h4\x22\x0d\x0a       clas\
s=\x22ColorScheme-T\
ext\x22\x0d\x0a       d=\x22\
m 5.918879,0.218\
75 v 15.875 l 10\
,-7.9375 z\x22\x0d\x0a   \
    style=\x22color\
:#232629;fill:cu\
rrentColor;fill-\
opacity:1;stroke\
:none;stroke-wid\
th:1.25996\x22 />\x0d\x0a\
  </g>\x0d\x0a</svg>\x0d\x0a\
\
\x00\x00\x0d\xad\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0d\x0a<svg\x0d\x0a   \
xmlns:dc=\x22http:/\
/purl.org/dc/ele\
ments/1.1/\x22\x0d\x0a   \
xmlns:cc=\x22http:/\
/creativecommons\
.org/ns#\x22\x0d\x0a   xm\
lns:rdf=\x22http://\
www.w3.org/1999/\
02/22-rdf-syntax\
-ns#\x22\x0d\x0a   xmlns:\
svg=\x22http://www.\
w3.org/2000/svg\x22\
\x0d\x0a   xmlns=\x22http\
://www.w3.org/20\
00/svg\x22\x0d\x0a   xmln\
s:sodipodi=\x22http\
://sodipodi.sour\
ceforge.net/DTD/\
sodipodi-0.dtd\x22\x0d\
\x0a   xmlns:inksca\
pe=\x22http://www.i\
nkscape.org/name\
spaces/inkscape\x22\
\x0d\x0a   viewBox=\x220 \
0 16 16\x22\x0d\x0a   ver\
sion=\x221.1\x22\x0d\x0a   i\
d=\x22svg6\x22\x0d\x0a   sod\
ipodi:docname=\x22e\
poch-encoder-und\
o.svg\x22\x0d\x0a   inksc\
ape:version=\x221.0\
 (4035a4fb49, 20\
20-05-01)\x22>\x0d\x0a  <\
metadata\x0d\x0a     i\
d=\x22metadata10\x22>\x0d\
\x0a    <rdf:RDF>\x0d\x0a\
      <cc:Work\x0d\x0a\
         rdf:abo\
ut=\x22\x22>\x0d\x0a        \
<dc:format>image\
/svg+xml</dc:for\
mat>\x0d\x0a        <d\
c:type\x0d\x0a        \
   rdf:resource=\
\x22http://purl.org\
/dc/dcmitype/Sti\
llImage\x22 />\x0d\x0a   \
     <dc:title><\
/dc:title>\x0d\x0a    \
  </cc:Work>\x0d\x0a  \
  </rdf:RDF>\x0d\x0a  \
</metadata>\x0d\x0a  <\
sodipodi:namedvi\
ew\x0d\x0a     inkscap\
e:document-rotat\
ion=\x220\x22\x0d\x0a     pa\
gecolor=\x22#ffffff\
\x22\x0d\x0a     borderco\
lor=\x22#666666\x22\x0d\x0a \
    borderopacit\
y=\x221\x22\x0d\x0a     obje\
cttolerance=\x2210\x22\
\x0d\x0a     gridtoler\
ance=\x2210\x22\x0d\x0a     \
guidetolerance=\x22\
10\x22\x0d\x0a     inksca\
pe:pageopacity=\x22\
0\x22\x0d\x0a     inkscap\
e:pageshadow=\x222\x22\
\x0d\x0a     inkscape:\
window-width=\x2219\
20\x22\x0d\x0a     inksca\
pe:window-height\
=\x221001\x22\x0d\x0a     id\
=\x22namedview8\x22\x0d\x0a \
    showgrid=\x22fa\
lse\x22\x0d\x0a     inksc\
ape:snap-bbox=\x22t\
rue\x22\x0d\x0a     inksc\
ape:snap-bbox-mi\
dpoints=\x22true\x22\x0d\x0a\
     inkscape:sn\
ap-bbox-edge-mid\
points=\x22true\x22\x0d\x0a \
    inkscape:bbo\
x-nodes=\x22true\x22\x0d\x0a\
     inkscape:bb\
ox-paths=\x22false\x22\
\x0d\x0a     inkscape:\
zoom=\x2232\x22\x0d\x0a     \
inkscape:cx=\x228.2\
26097\x22\x0d\x0a     ink\
scape:cy=\x227.5297\
539\x22\x0d\x0a     inksc\
ape:window-x=\x22-9\
\x22\x0d\x0a     inkscape\
:window-y=\x2241\x22\x0d\x0a\
     inkscape:wi\
ndow-maximized=\x22\
1\x22\x0d\x0a     inkscap\
e:current-layer=\
\x22g885\x22 />\x0d\x0a  <de\
fs\x0d\x0a     id=\x22def\
s3051\x22>\x0d\x0a    <st\
yle\x0d\x0a       type\
=\x22text/css\x22\x0d\x0a   \
    id=\x22current-\
color-scheme\x22>\x0d\x0a\
      .ColorSche\
me-Text {\x0d\x0a     \
   color:#232629\
;\x0d\x0a      }\x0d\x0a    \
  </style>\x0d\x0a  </\
defs>\x0d\x0a  <g\x0d\x0a   \
  id=\x22g40\x22\x0d\x0a    \
 transform=\x22tran\
slate(0.03124978\
,-0.15625)\x22>\x0d\x0a  \
  <g\x0d\x0a       tra\
nsform=\x22matrix(-\
1.0478301,0,0,1.\
0478301,15.30897\
,-0.87676339)\x22\x0d\x0a\
       id=\x22g871\x22\
>\x0d\x0a      <g\x0d\x0a   \
      transform=\
\x22translate(-0.41\
834834,-0.121797\
88)\x22\x0d\x0a         i\
d=\x22g881\x22>\x0d\x0a     \
   <g\x0d\x0a         \
  transform=\x22mat\
rix(-1,0,0,1,14.\
891757,0.0596470\
5)\x22\x0d\x0a           \
id=\x22g885\x22>\x0d\x0a    \
      <g\x0d\x0a      \
       transform\
=\x22matrix(-1,0,0,\
1,14.93633,1.628\
9895e-4)\x22\x0d\x0a     \
        id=\x22g889\
\x22>\x0d\x0a            \
<path\x0d\x0a         \
      id=\x22path4-\
6\x22\x0d\x0a            \
   class=\x22ColorS\
cheme-Text\x22\x0d\x0a   \
            d=\x22M\
 10.056273,1.354\
6116 7.8826465,8\
.3895922 13.4009\
44,6.2413157 Z\x22\x0d\
\x0a               \
style=\x22color:#23\
2629;fill:curren\
tColor;fill-opac\
ity:1;stroke:non\
e;stroke-width:0\
.584393\x22 />\x0d\x0a   \
         <path\x0d\x0a\
               s\
tyle=\x22color:#000\
000;overflow:vis\
ible;fill:none;s\
troke:#232629;st\
roke-width:1.963\
1;stroke-linecap\
:square;stroke-m\
iterlimit:4;stro\
ke-dasharray:non\
e;stroke-opacity\
:1\x22\x0d\x0a           \
    id=\x22path837-\
5\x22\x0d\x0a            \
   transform=\x22ro\
tate(49.43912)\x22\x0d\
\x0a               \
d=\x22M 17.566429,0\
.6638031 A 5.061\
4376,5.1534634 0\
 0 1 13.194504,5\
.7692229 5.06143\
76,5.1534634 0 0\
 1 7.6314166,2.0\
548116 5.0614376\
,5.1534634 0 0 1\
 10.487638,-4.06\
26264\x22\x0d\x0a        \
       sodipodi:\
type=\x22arc\x22\x0d\x0a    \
           sodip\
odi:arc-type=\x22ar\
c\x22\x0d\x0a            \
   sodipodi:star\
t=\x220\x22\x0d\x0a         \
      sodipodi:e\
nd=\x224.3024284\x22\x0d\x0a\
               s\
odipodi:ry=\x225.15\
34634\x22\x0d\x0a        \
       sodipodi:\
rx=\x225.0614376\x22\x0d\x0a\
               s\
odipodi:cy=\x220.66\
38031\x22\x0d\x0a        \
       sodipodi:\
cx=\x2212.504992\x22\x0d\x0a\
               s\
odipodi:open=\x22tr\
ue\x22 />\x0d\x0a        \
  </g>\x0d\x0a        \
</g>\x0d\x0a      </g>\
\x0d\x0a    </g>\x0d\x0a  </\
g>\x0d\x0a</svg>\x0d\x0a\
\x00\x00\x01p\
<\
svg xmlns=\x22http:\
//www.w3.org/200\
0/svg\x22 viewBox=\x22\
0 0 16 16\x22>\x0d\x0a  <\
defs id=\x22defs305\
1\x22>\x0d\x0a    <style \
type=\x22text/css\x22 \
id=\x22current-colo\
r-scheme\x22>\x0d\x0a    \
  .ColorScheme-T\
ext {\x0d\x0a        c\
olor:#232629;\x0d\x0a \
     }\x0d\x0a      </\
style>\x0d\x0a  </defs\
>\x0d\x0a <path style=\
\x22fill:currentCol\
or;fill-opacity:\
1;stroke:none\x22\x0d\x0a\
       d=\x22m 3 3 \
0 10 10 -5 z\x22\x0d\x0a \
    class=\x22Color\
Scheme-Text\x22\x0d\x0a  \
   />\x0d\x0a</svg>\x0d\x0a\
\x00\x00\x01u\
<\
svg xmlns=\x22http:\
//www.w3.org/200\
0/svg\x22 viewBox=\x22\
0 0 16 16\x22>\x0d\x0a  <\
defs id=\x22defs305\
1\x22>\x0d\x0a    <style \
type=\x22text/css\x22 \
id=\x22current-colo\
r-scheme\x22>\x0d\x0a    \
  .ColorScheme-T\
ext {\x0d\x0a        c\
olor:#232629;\x0d\x0a \
     }\x0d\x0a      </\
style>\x0d\x0a  </defs\
>\x0d\x0a <path style=\
\x22fill:currentCol\
or;fill-opacity:\
1;stroke:none\x22\x0d\x0a\
       d=\x22m 3 3 \
0 10 10 0 0 -10 \
z\x22\x0d\x0a     class=\x22\
ColorScheme-Text\
\x22\x0d\x0a     />\x0d\x0a</sv\
g>\x0d\x0a\
\x00\x00\x0d\xab\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0d\x0a<svg\x0d\x0a   \
xmlns:dc=\x22http:/\
/purl.org/dc/ele\
ments/1.1/\x22\x0d\x0a   \
xmlns:cc=\x22http:/\
/creativecommons\
.org/ns#\x22\x0d\x0a   xm\
lns:rdf=\x22http://\
www.w3.org/1999/\
02/22-rdf-syntax\
-ns#\x22\x0d\x0a   xmlns:\
svg=\x22http://www.\
w3.org/2000/svg\x22\
\x0d\x0a   xmlns=\x22http\
://www.w3.org/20\
00/svg\x22\x0d\x0a   xmln\
s:sodipodi=\x22http\
://sodipodi.sour\
ceforge.net/DTD/\
sodipodi-0.dtd\x22\x0d\
\x0a   xmlns:inksca\
pe=\x22http://www.i\
nkscape.org/name\
spaces/inkscape\x22\
\x0d\x0a   viewBox=\x220 \
0 16 16\x22\x0d\x0a   ver\
sion=\x221.1\x22\x0d\x0a   i\
d=\x22svg6\x22\x0d\x0a   sod\
ipodi:docname=\x22e\
poch-encoder-red\
o.svg\x22\x0d\x0a   inksc\
ape:version=\x221.0\
 (4035a4fb49, 20\
20-05-01)\x22>\x0d\x0a  <\
metadata\x0d\x0a     i\
d=\x22metadata10\x22>\x0d\
\x0a    <rdf:RDF>\x0d\x0a\
      <cc:Work\x0d\x0a\
         rdf:abo\
ut=\x22\x22>\x0d\x0a        \
<dc:format>image\
/svg+xml</dc:for\
mat>\x0d\x0a        <d\
c:type\x0d\x0a        \
   rdf:resource=\
\x22http://purl.org\
/dc/dcmitype/Sti\
llImage\x22 />\x0d\x0a   \
     <dc:title><\
/dc:title>\x0d\x0a    \
  </cc:Work>\x0d\x0a  \
  </rdf:RDF>\x0d\x0a  \
</metadata>\x0d\x0a  <\
sodipodi:namedvi\
ew\x0d\x0a     inkscap\
e:document-rotat\
ion=\x220\x22\x0d\x0a     pa\
gecolor=\x22#ffffff\
\x22\x0d\x0a     borderco\
lor=\x22#666666\x22\x0d\x0a \
    borderopacit\
y=\x221\x22\x0d\x0a     obje\
cttolerance=\x2210\x22\
\x0d\x0a     gridtoler\
ance=\x2210\x22\x0d\x0a     \
guidetolerance=\x22\
10\x22\x0d\x0a     inksca\
pe:pageopacity=\x22\
0\x22\x0d\x0a     inkscap\
e:pageshadow=\x222\x22\
\x0d\x0a     inkscape:\
window-width=\x2219\
20\x22\x0d\x0a     inksca\
pe:window-height\
=\x221001\x22\x0d\x0a     id\
=\x22namedview8\x22\x0d\x0a \
    showgrid=\x22fa\
lse\x22\x0d\x0a     inksc\
ape:snap-bbox=\x22t\
rue\x22\x0d\x0a     inksc\
ape:snap-bbox-mi\
dpoints=\x22true\x22\x0d\x0a\
     inkscape:sn\
ap-bbox-edge-mid\
points=\x22true\x22\x0d\x0a \
    inkscape:bbo\
x-nodes=\x22true\x22\x0d\x0a\
     inkscape:bb\
ox-paths=\x22false\x22\
\x0d\x0a     inkscape:\
zoom=\x2232\x22\x0d\x0a     \
inkscape:cx=\x228.2\
26097\x22\x0d\x0a     ink\
scape:cy=\x227.5297\
539\x22\x0d\x0a     inksc\
ape:window-x=\x22-9\
\x22\x0d\x0a     inkscape\
:window-y=\x2241\x22\x0d\x0a\
     inkscape:wi\
ndow-maximized=\x22\
1\x22\x0d\x0a     inkscap\
e:current-layer=\
\x22g885\x22 />\x0d\x0a  <de\
fs\x0d\x0a     id=\x22def\
s3051\x22>\x0d\x0a    <st\
yle\x0d\x0a       type\
=\x22text/css\x22\x0d\x0a   \
    id=\x22current-\
color-scheme\x22>\x0d\x0a\
      .ColorSche\
me-Text {\x0d\x0a     \
   color:#232629\
;\x0d\x0a      }\x0d\x0a    \
  </style>\x0d\x0a  </\
defs>\x0d\x0a  <g\x0d\x0a   \
  id=\x22g40\x22\x0d\x0a    \
 transform=\x22tran\
slate(0.03124978\
,-0.15625)\x22>\x0d\x0a  \
  <g\x0d\x0a       tra\
nsform=\x22matrix(-\
1.0478301,0,0,1.\
0478301,15.30897\
,-0.87676339)\x22\x0d\x0a\
       id=\x22g871\x22\
>\x0d\x0a      <g\x0d\x0a   \
      transform=\
\x22translate(-0.41\
834834,-0.121797\
88)\x22\x0d\x0a         i\
d=\x22g881\x22>\x0d\x0a     \
   <g\x0d\x0a         \
  transform=\x22mat\
rix(-1,0,0,1,14.\
891757,0.0596470\
5)\x22\x0d\x0a           \
id=\x22g885\x22>\x0d\x0a    \
      <g\x0d\x0a      \
       transform\
=\x22translate(1.62\
53609e-4,1.62898\
95e-4)\x22\x0d\x0a       \
      id=\x22g889\x22>\
\x0d\x0a            <p\
ath\x0d\x0a           \
    id=\x22path4-6\x22\
\x0d\x0a              \
 class=\x22ColorSch\
eme-Text\x22\x0d\x0a     \
          d=\x22M 1\
0.056273,1.35461\
16 7.8826465,8.3\
895922 13.400944\
,6.2413157 Z\x22\x0d\x0a \
              st\
yle=\x22color:#2326\
29;fill:currentC\
olor;fill-opacit\
y:1;stroke:none;\
stroke-width:0.5\
84393\x22 />\x0d\x0a     \
       <path\x0d\x0a  \
             sty\
le=\x22color:#00000\
0;overflow:visib\
le;fill:none;str\
oke:#232629;stro\
ke-width:1.9631;\
stroke-linecap:s\
quare;stroke-mit\
erlimit:4;stroke\
-dasharray:none;\
stroke-opacity:1\
\x22\x0d\x0a             \
  id=\x22path837-5\x22\
\x0d\x0a              \
 transform=\x22rota\
te(49.43912)\x22\x0d\x0a \
              d=\
\x22M 17.566429,0.6\
638031 A 5.06143\
76,5.1534634 0 0\
 1 13.194504,5.7\
692229 5.0614376\
,5.1534634 0 0 1\
 7.6314166,2.054\
8116 5.0614376,5\
.1534634 0 0 1 1\
0.487638,-4.0626\
264\x22\x0d\x0a          \
     sodipodi:ty\
pe=\x22arc\x22\x0d\x0a      \
         sodipod\
i:arc-type=\x22arc\x22\
\x0d\x0a              \
 sodipodi:start=\
\x220\x22\x0d\x0a           \
    sodipodi:end\
=\x224.3024284\x22\x0d\x0a  \
             sod\
ipodi:ry=\x225.1534\
634\x22\x0d\x0a          \
     sodipodi:rx\
=\x225.0614376\x22\x0d\x0a  \
             sod\
ipodi:cy=\x220.6638\
031\x22\x0d\x0a          \
     sodipodi:cx\
=\x2212.504992\x22\x0d\x0a  \
             sod\
ipodi:open=\x22true\
\x22 />\x0d\x0a          \
</g>\x0d\x0a        </\
g>\x0d\x0a      </g>\x0d\x0a\
    </g>\x0d\x0a  </g>\
\x0d\x0a</svg>\x0d\x0a\
\x00\x00\x09\xb5\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0d\x0a<svg\x0d\x0a   \
xmlns:dc=\x22http:/\
/purl.org/dc/ele\
ments/1.1/\x22\x0d\x0a   \
xmlns:cc=\x22http:/\
/creativecommons\
.org/ns#\x22\x0d\x0a   xm\
lns:rdf=\x22http://\
www.w3.org/1999/\
02/22-rdf-syntax\
-ns#\x22\x0d\x0a   xmlns:\
svg=\x22http://www.\
w3.org/2000/svg\x22\
\x0d\x0a   xmlns=\x22http\
://www.w3.org/20\
00/svg\x22\x0d\x0a   xmln\
s:sodipodi=\x22http\
://sodipodi.sour\
ceforge.net/DTD/\
sodipodi-0.dtd\x22\x0d\
\x0a   xmlns:inksca\
pe=\x22http://www.i\
nkscape.org/name\
spaces/inkscape\x22\
\x0d\x0a   viewBox=\x220 \
0 16 16\x22\x0d\x0a   ver\
sion=\x221.1\x22\x0d\x0a   i\
d=\x22svg6\x22\x0d\x0a   sod\
ipodi:docname=\x22e\
poch-encoder-del\
ete.svg\x22\x0d\x0a   ink\
scape:version=\x221\
.0 (4035a4fb49, \
2020-05-01)\x22>\x0d\x0a \
 <metadata\x0d\x0a    \
 id=\x22metadata10\x22\
>\x0d\x0a    <rdf:RDF>\
\x0d\x0a      <cc:Work\
\x0d\x0a         rdf:a\
bout=\x22\x22>\x0d\x0a      \
  <dc:format>ima\
ge/svg+xml</dc:f\
ormat>\x0d\x0a        \
<dc:type\x0d\x0a      \
     rdf:resourc\
e=\x22http://purl.o\
rg/dc/dcmitype/S\
tillImage\x22 />\x0d\x0a \
       <dc:title\
 />\x0d\x0a      </cc:\
Work>\x0d\x0a    </rdf\
:RDF>\x0d\x0a  </metad\
ata>\x0d\x0a  <sodipod\
i:namedview\x0d\x0a   \
  inkscape:docum\
ent-rotation=\x220\x22\
\x0d\x0a     pagecolor\
=\x22#ffffff\x22\x0d\x0a    \
 bordercolor=\x22#6\
66666\x22\x0d\x0a     bor\
deropacity=\x221\x22\x0d\x0a\
     objecttoler\
ance=\x2210\x22\x0d\x0a     \
gridtolerance=\x221\
0\x22\x0d\x0a     guideto\
lerance=\x2210\x22\x0d\x0a  \
   inkscape:page\
opacity=\x220\x22\x0d\x0a   \
  inkscape:pages\
hadow=\x222\x22\x0d\x0a     \
inkscape:window-\
width=\x221920\x22\x0d\x0a  \
   inkscape:wind\
ow-height=\x221017\x22\
\x0d\x0a     id=\x22named\
view8\x22\x0d\x0a     sho\
wgrid=\x22false\x22\x0d\x0a \
    inkscape:sna\
p-bbox=\x22true\x22\x0d\x0a \
    inkscape:sna\
p-bbox-midpoints\
=\x22true\x22\x0d\x0a     in\
kscape:snap-bbox\
-edge-midpoints=\
\x22true\x22\x0d\x0a     ink\
scape:bbox-nodes\
=\x22true\x22\x0d\x0a     in\
kscape:zoom=\x2232.\
482718\x22\x0d\x0a     in\
kscape:cx=\x228.707\
8372\x22\x0d\x0a     inks\
cape:cy=\x228.35135\
5\x22\x0d\x0a     inkscap\
e:window-x=\x22-8\x22\x0d\
\x0a     inkscape:w\
indow-y=\x2232\x22\x0d\x0a  \
   inkscape:wind\
ow-maximized=\x221\x22\
\x0d\x0a     inkscape:\
current-layer=\x22s\
vg6\x22 />\x0d\x0a  <defs\
\x0d\x0a     id=\x22defs3\
051\x22>\x0d\x0a    <styl\
e\x0d\x0a       type=\x22\
text/css\x22\x0d\x0a     \
  id=\x22current-co\
lor-scheme\x22>\x0d\x0a  \
    .ColorScheme\
-Text {\x0d\x0a       \
 color:#232629;\x0d\
\x0a      }\x0d\x0a      \
</style>\x0d\x0a  </de\
fs>\x0d\x0a  <g\x0d\x0a     \
id=\x22g33\x22\x0d\x0a     t\
ransform=\x22matrix\
(1.7776366,0,0,1\
.7776366,14.1839\
94,-7.7883383)\x22>\
\x0d\x0a    <rect\x0d\x0a   \
    transform=\x22r\
otate(45)\x22\x0d\x0a    \
   y=\x223.5835483\x22\
\x0d\x0a       x=\x222.95\
8411\x22\x0d\x0a       he\
ight=\x2210.313176\x22\
\x0d\x0a       width=\x22\
1.7239937\x22\x0d\x0a    \
   id=\x22rect12\x22\x0d\x0a\
       style=\x22co\
lor:#000000;over\
flow:visible;fil\
l:#232629;stroke\
-width:12.3402;s\
troke-linecap:ro\
und\x22 />\x0d\x0a    <re\
ct\x0d\x0a       trans\
form=\x22rotate(135\
)\x22\x0d\x0a       style\
=\x22color:#000000;\
overflow:visible\
;fill:#232629;st\
roke-width:12.34\
02;stroke-lineca\
p:round\x22\x0d\x0a      \
 id=\x22rect12-7\x22\x0d\x0a\
       width=\x221.\
7239937\x22\x0d\x0a      \
 height=\x2210.3131\
76\x22\x0d\x0a       x=\x227\
.878139\x22\x0d\x0a      \
 y=\x22-8.9769955\x22 \
/>\x0d\x0a  </g>\x0d\x0a</sv\
g>\x0d\x0a\
\x00\x00\x05\x13\
\x00\
\x00\x13Tx\x9c\xd5X\xdbn\xe36\x10}_`\xff\
\xc1P^\x124\xa4H\x8a\xba9v\x16h\x17\x0b\xf4\
\xb5\x9b\xa2\xcf\xb2D\xcbj$\xd1\x10\xe9\xd8\xd9E\xff\
\xbdC\xc9\x92eKN\x9c\xa6\x17T\x89\x11\x91sf\
H\xce\x1c\x9e12\xfb\xb4+\xf2\xc9\x93\xa8T&\xcb\
\xb9E1\xb1&\xa2\x8ce\x92\x95\xe9\xdc\xfa\xf5\xe1\x0b\
\x0a\xac\x89\xd2Q\x99D\xb9,\xc5\xdc*\xa5\xf5\xe9\xfe\
\xe3\x87\x99zJ?~\x98L&\xe0^\xaai\x12\xcf\
\xad\x95\xd6\xeb\xa9m\xaf7U\x8ee\x95\xdaIl\x8b\
\x5c\x14\xa2\xd4\xca\xa6\x98\xdaV\x0f\x1f\x1f\xf0q%\x22\
\x9d=\x89X\x16\x85,U\xedZ\xaa\xab>\xbaJ\x96\
\x1d|\xbb\xdd\xe2\xadS\xa3h\x18\x866a6c\x08\
\x10H=\x97:\xda\xa1\x13_\xd8\xe7\x98/#\x84\xd8\
`\xebA/\x84M\x15$g\x0d\x9f\x0e\xdfN`%\
7U,\x96\xe0(p)\xb4\xfd\xf9\xe1sgD\x04\
':\xe9\xc7\xc9\xcaG\x15Gkq\xb4n;\xd9\xa4\
!*\x84ZG\xb1Pv;\xdf\x04hG\xd3~\xe1\
&\xd7\x9c8n\xc4\x97\x0b\x1e\xdeN\x18a\x04\x11\x17\
\x11z\xd3\xf8\xb4\x1b\x99&26\x81\xe7\x96X\xcbx\
\x85\xeaj\x8b\x0a%\x9bu\x9e\xc5\x91\x16\xb8;o\x96\
\xcc-\x18x\xcd\xa8\xb7\x16\xdd\xcfdb\xfb\xa3\xdc\xcd\
-2!\x13\xea\xc1\xafuo\x0c\xb3B\xe8(\x89t\
T\xa3\x9a8\xed\x14%\x0d\x06PP\xb4\xe9/\x9f\xbf\
\xec\x870\x11\xc7\xd3\xdfd\xf5\xd8\x8e\xe11\x90h!\
7znY\xf7\x87\xf9Y\x12O!\xcdE\xa4\xef\xb3\
\x22J\x85)\xd1\x0f\x90\xd5\x99}0\x1c\xa3\xf5\xf3Z\
\xf4\xe26\x91+\xd1Tl\x94\xbbI\x5cd\xc6\xcb\xfe\
\xaa\xb3<\xff\xd9,cM\xec\xd3\xb0\x99\xce\xc5}\xbd\
l\xf3\xda\x9d\xc5\xde\x1f\xa6=\xac\xdd?\xed\xccn\xd3\
\xd1\x0c\xbb\xda\x98\xc2$&\xadm\xe6\xdaB\xc7\x9b\xaa\
\x82\x9b\x84\xf2\xe8YTs+\xa5\xe1\xbe*=\xcc6\
+\x13\xb9EE\xb4\xcb\x8a\xec\x9b\x80\xa4\xd3s\x98\xe7\
\xb9\xc5\xcf\x1a\xa1\xa0(\x1c\x18c\xf0A\x0c\x13\xcf\xf3\
\x02\x87\x0f\xcd\xe0E\x03\xec\xf8!\xf7\xd9\xc0\xfaM\xca\
\x02\xec\x1ef\x9c:\xee0\xf8b!wh\x1d\xe9\x15\
\xdcB]m\xc48\xa0\x04\xa2\x9e\x03\xa82Z\xa3\x1a\
%\x92T\xa0\x22K\xd62\x03\xe9y\x15~1\xf2\xd8\
\xaeVr\x9bV\x86\xd9\xcb(W\x077\x98\xe8\x8a\x18\
\x9c\xcb\xf0Jd\xe9\x0a8M\x099[\x85m\x96\xe8\
\x15@BF\x06\x905\x90Q\xad\x22\x80\xcd\xada\xb2\
\x8dU\x82l\x00}\xe1j\xb6\xe6t\x93%B\xcb\x5c\
TQi(O\x0f\x168\xc6\xa8A.~\x17\xb1\x1e\
5-d\x05\xa2\xd1\xadB\x8f\xe7c\x99K`\xe9\x95\
W?\xad\xcd\xec\xab\xb5,\xebg\xb0u\xd0\xa6\x8d\xe9\
\x18\xa8\x92\x1a\xba\x82Q\x9b\xe1\xe9\xeb\x8a\xa4\xb9\x5cD\
y\x9b\xfd\xfd\xc5\x9c%b\xa9z\x950C\x87\xb8\xb4\
\x93\x1c\xa5\x9f\xf3\x83\x0e\x18H{\xb1\xea\x8d!\x15\xaf\
\xa0cY\x1d\xc2\x08\x00\xd4]\xec\xb4\x1d+u\x90 \
\xfc\x93\x81\x7f\xad\xd1\xe8\x01\xcc\x93\xef\x07a\xa8CM\
\xaf\x98\xc3<\x16\xde\xb5\xf3\x7f\x1c\xa4\xa1\xde\xc5^\x08\
\xcc\x16\x9b\xd7\xb4\xb7\xf1\xd4\xed\xb8\xa3!\xf7\xcah\x9a\
\xe1\x1f\xbc\xe6 \xd1\xd7\x04\x13\x97sF=~\x0b\x8d\
\xc5\xa1\x0eg\xc4g7\xdd9\xd3\xa33\xf6\xd5\xe2l\
@h\x18\x9e\x7fK\xb0\xe7\x90\xd0\xf5\xc2C\xb0\xa3p\
m@\x16\x1e\xe9q\x1f\xd0B\x5c\xdf:\x9a\xed\xad\x0b\
\xfa\x5ce\xbbkDo\x09\xfc\xd0[\xeab\x9f\xfb\x84\
Ss\x1a\x971\xce\x83\xd0\xbd\xe9\xaf`\xda\x05p\xf1\
(\xe0Q\xc8\x9a0\xe2:$7\xd6\x09\xc8\xe8\x16\xc5\
!a\x94;\xec\xd4hT\x0b{A\xe8\x05\x94\x9e\xda\
\xba[\x8a\xdd\xc0\x09\xc9\xd0y\x7fEA\xf5\x5c\x12x\
\xcc=\xb5\x9b<\x98]S\x8e|4\xb0\xd6,\x00\x02\
6l!\xf5s'\xa1\xcb.s\xb9\x9d>e*[\
\xe4\xe2n\x09\xcd\xa7\xe3\x92\xd2\x95|\x14\x8d4L\x19\
\xc5\xcc!^;\x99g\xa5\x80\xeb1\xad\xe4\xa6L\x8e\
;\xd5x\xf2\xfe\xd1\xf5_\xcc\x04\xf2\xdf\x9a\xc9WK\
\xf1R\x1dk\x020\xecp7\x1c\xd8\xc6\x19tA\xf6\
z\x8e\xa0J\xb9\x002\x8f3\x0f\xae*7\x97ud\
\xc7gwu)\xf3\xc0\x9f0\xdf\x1b\xa4\xf34\xddh\
\xb0\xc2\xbb\x8b\xefb\xee\x87\xce\x7fG\xbe\x97\xd6\x7f5\
\x1b\xc8{cB/\xe1\xdf\xd9j\x1a\x1a8\x98\xf9\x8e\
\x1b\xf0\x97\x08x\xe0\xd1\xf1\xf7L;}]jO+\
<*\xf2\x0e&\x01\x0bBJo9\x06\x1f\xea\xb2\x0b\
4\xf6\xbd\x95\xa2>\xa6!\xffK\x95\x0a\xcfT\xc9\xd4\
\xdew\xc2a\x15.)\x13\xc7\xae\xe7P\x97\x0c.$\
\x94\xc9\xc7\x9c\xfa\x90\xa0\x81\xfe\x0c{W\xdd\xb6\xea\xf6\
\xf5f\xbd\x18\x8b0J\x19\xc2\xdc\xc0w\x07\x5c\xfd\x17\
{\x16|\x06w\xe1\x7f\xdf\xb7\xde'\x1d\x03\xde\xfc\x1d\
\xd2q\xb6\xd6/\xb6\x90\x81v\xc0\xf7\xa7S\xf5\xb8\xb0\
{5\x9eo\x11\xae\x17\xb7})\x0d9&\xd4\xf3\x83\
a\x801\xcd\x1e\xea\xc1\xbb\xb5\x89c\x9f\x9d\xd3\xa6\xb3\
\x22|xo\xdf\x9a\xbf3\xf3o\x08\xf8\xfb'\xc4*\
\xeb|\
\x00\x00\x04\x95\
\x00\
\x00\x10Qx\x9c\xd5W\xdbn\xe36\x10}_`\xff\
AP^\x124\xa4H\xeabK\xb1\xb3@\x1b,\xd0\
\xd7n\x8a>\xcb\x12-\xab\x91DA\xa4c{\x17\xfd\
\xf7\x0e\xa5P\x17_\x9a\x04\x08\xd0V\x8eaq\xe6\x90\
\x1c\x9e93R\x16_\xf6ea=\xf3F\xe6\xa2Z\
\xda\x14\x13\xdb\xe2U\x22\xd2\xbc\xca\x96\xf6\xef\x8f_\xd1\
\xdc\xb6\xa4\x8a\xab4.D\xc5\x97v%\xec/\xf7\x9f\
?-\xe4s\xf6\xf9\x93eY0\xbd\x92Q\x9a,\xed\
\x8dRu\xe48\xf5\xb6)\xb0h2'M\x1c^\xf0\
\x92WJ:\x14S\xc7\x1e\xe1\x93\x01\x9f4<V\xf9\
3ODY\x8aJ\xb6S+y5F7\xe9\xba\x87\
\xefv;\xbcs[\x14\x0d\xc3\xd0!\xcca\x0c\x01\x02\
\xc9C\xa5\xe2=:\x9a\x0bq\x9e\x9b\xcb\x08!\x0e\xf8\
F\xd07\xc2\x22\x09\xe4\xd4\xf0\xed\xf1\xc6\x80\xa5\xd86\
\x09_\xc3D\x8e+\xae\x9c\x87\xc7\x87\xde\x89\x08NU\
:^'\xaf\x9ed\x12\xd7|\xb2\xaf1v4\xc4%\
\x97u\x9cp\xe9\x18{\xb7\x80\x19E\xe3\xc4Y\xd7\x1e\
q\xfd\xd8[\xaf\xbc\xf0\xd6b\x84\x11D|D\xe8M\
7\xc7\x04\x12\xa5\x22\xd1\x0b/m^\x8bd\x83\xdal\
\xf3\x06\xc9\xba\xc8\x15\xee\xcf\x9a\xa7K\x1b\x06A7\x1a\
\xedC_,9\xdf\xfd,\xf6K\x9bX\xc4\xa2\x01\xfc\
\xd9\xf7\xda\xb1(\xb9\x8a\xd3X\xc5-\xaa[\xc7\x98(\
\xe90\x80\x82\x84E\xbf=|}\x19\x82!I\xa2?\
D\xf3d\xc6piH\xbc\x12[\xb5\xb4\xed\xfb\xc1\xbe\
H\x93\x08(.cu\x9f\x97q\xc6uz~\x02F\
\x17\xce\xe0\x98\xa2\xd5\xa1\xe6\xa3u\xbb\x95\x1b\xdee\xeb\
\xacn\xd3\xa4\xcc\xf5,\xe7\x9b\xca\x8b\xe2W\xbd\x8dm\
9C\xb0\xceK\xb4\xe64\xce\xf88\x0b\xc7\x9c\xb7\x1b\
\xf6\xc4k\xd6S\xcd\x9b\xa1\xc6d1\xd96\x0d\x94\x09\
*\xe2\x03o\x96vF]\xcf>\xc6\xec\xf2*\x15;\
T\xc6\xfb\xbc\xcc\xbfs`\x95^\xc2\x1c\x96\xb6\xcb.\
9!cP\xd2'\x11\xc0\x9c\x19\xa6^@A7\xa7\
\xde\xbdN\xbcK\x19%Ap\xe2\xfd.Dyv\xc7\
\xd5J\xecQ\x1d\xab\x0d\x94\xd6:.$?\x8f\xa8@\
~\x80P\xcd\xf6\x14 \xab\xb8F-\x8a\xa7\x19Ge\
\x9e\xd6\x22\x87\x86\xf2*\xfc\xcd\xc8\xa9_n\xc4.k\
\xb4f\xa7\x01\x83\xa1\xcf\xde){/\xd4nx\x9em\
@\xad\x94\xd0\xd9%\xcc.O\xd5\x06 !#'\x90\
\x1ad&71\xc0\x96\xf6)\x9b\xda+\xa0\x19\x800\
\xa1\xe8\x8c;\xdb\xe6)W\xa2\xe0M\x5ci1\xd3\xc1\
\x03\xc78\xeb\x10\xab?y\xa2\xce\xbaV\xa2\x81V\xd0\
\xefB\xa7\xf6D\x14\x02\xe4y\x15\xb4\x97\xf1\xe9\xb8\x8c\
g\xdd^'\xa1C\xc7\xd9\xea\xe7\x00j\x84\x82^\xaf\
\xfb\x081\xf5\xb4H\xf9Z\x8eh\xd6C\x97\xf8\xb4\xef\
\x14R\x1d\x8a\xa1|5\xc4\x94K\xbb+\x92\xc9\x06\x1e\
2v\x8f\xd0u\x0bI\xe5{\xe5$R\x0e\x9d\x03\xff\
\xa2\xe1\xdfZ4z\x04\xb7\xf5ch\x0a\xedR\xd1\x15\
sY\xc0\xc2;c\xffk(\xf86\x8a\x97\xf2\xd6!\
v\xb7\xd9(\xf0\xb6jM\xd0\xd9\x10\x0e\x90,u[\
\xd2B\x83\xdb\x22V\xfc\x1a\x1e\x06\x04\xca\xc9'\xfe-\
\xdc\x05\xcc\xbf\xb1''\xcc(!C\xe4\x8b\x06\xf25\
j`m,@C\x173i\xaf;\x01-z]\x88\
]\xf4\x9c\xcb|U\xf0\xbb5t\xae\xfeDR5\xe2\
\x89w\xea\x8bh\x80\x03\xea\x19[\x91W\x1c\xb2\x145\
b[\xa5\xf6h\x1b\x1d\x88\xde\x992\x14\x8c\xedF\xc2\
\xd8\xa7t\xcef\xb3\xb1\xaf/\x01\x1f{\x01\x9b\xf9\x13\
\xe7^\xf7\x180\xban\xe0\x8e\xedZ\xcf\x98\x11\xe8=\
\xd4\xf3&}6;\x0a'\xa3\x01C\x93\xb9g\xe9\xa5\
\x04\xcf]w\x0e\xec\x22C\xef\xf8\x91pD\xe7\x07\x10\
\xeab\x88\xdd}\x95\xd1\x09\xa7\x04\x85\xc1\xd4yD\xec\
|\xea5\xd4\x868\xf0\xbc\x99\xcf\xfc\xa9\x1b\xc8u\xb1\
O\xa8\xcb\xfc\xa3e\xf5\xe3\x00\xd3\xd9,`\xeel\xcc\
\xee9&\x00\x8b|\x0c\xabP\xdf\x0d\xcel`\x969\
\x1b\x99\x8f\xfd`\x1eR\x7f\xfe\x9ecM\x18\x81\x8f\xfd\
\xa1\x99\x09q8#^\xf0\x86\xcc\x8c\x84\xd4v)~\
\x1d\x92\x9b\xd7\xe8:?\xe9_:\xc1\x11\x93\xf3c.\
\xdf&\xaf\x0bI\xd4o\x00\xfa\x15\x80\xf8\xde\xfcD^\
#\xc9L\xde\x92\xb2W*\xd9\x9b6\x96\x11\x9b\xf0\x12\
\xd7\xe4\xfbkDo\x09|\xe8\xad\x8f\x19\x9b\x07a\xf0\
\xc6\x82>\x5c\xd4\xea\xe52y\xa5\xbe\xde\xa1\xe2\x10\xd1\
\x8f\xd5\xc0?\xf6\x97\xff\xb1D\xdd\x8f\x96\xe8\x85\xa4_\
\x94\xe8\xfb[\xe0\x85\x22\xf8\xc8\x168G\xfe\xf4\x19\xf7\
\x9fj\x83CU\x9b\xbb\xeew\xa1\xff\x0f\x83\xdf\xbf\x01\
\xa1\x0cC\xcd\
\x00\x00\x0d\xa5\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0d\x0a<svg\x0d\x0a   \
xmlns:dc=\x22http:/\
/purl.org/dc/ele\
ments/1.1/\x22\x0d\x0a   \
xmlns:cc=\x22http:/\
/creativecommons\
.org/ns#\x22\x0d\x0a   xm\
lns:rdf=\x22http://\
www.w3.org/1999/\
02/22-rdf-syntax\
-ns#\x22\x0d\x0a   xmlns:\
svg=\x22http://www.\
w3.org/2000/svg\x22\
\x0d\x0a   xmlns=\x22http\
://www.w3.org/20\
00/svg\x22\x0d\x0a   xmln\
s:sodipodi=\x22http\
://sodipodi.sour\
ceforge.net/DTD/\
sodipodi-0.dtd\x22\x0d\
\x0a   xmlns:inksca\
pe=\x22http://www.i\
nkscape.org/name\
spaces/inkscape\x22\
\x0d\x0a   inkscape:ex\
port-ydpi=\x22900\x22\x0d\
\x0a   inkscape:exp\
ort-xdpi=\x22900\x22\x0d\x0a\
   inkscape:expo\
rt-filename=\x22C:\x5c\
Users\x5cJeffrey\x5cDo\
cuments\x5cpython-d\
ev\x5cephyviewer\x5cgi\
t-repo\x5cephyviewe\
r\x5cephyviewer\x5cico\
ns\x5cimg\x5cepoch-enc\
oder-save.png\x22\x0d\x0a\
   inkscape:vers\
ion=\x221.0 (4035a4\
fb49, 2020-05-01\
)\x22\x0d\x0a   sodipodi:\
docname=\x22epoch-e\
ncoder-save.svg\x22\
\x0d\x0a   id=\x22svg6\x22\x0d\x0a\
   version=\x221.1\x22\
\x0d\x0a   viewBox=\x220 \
0 16 16\x22>\x0d\x0a  <me\
tadata\x0d\x0a     id=\
\x22metadata10\x22>\x0d\x0a \
   <rdf:RDF>\x0d\x0a  \
    <cc:Work\x0d\x0a  \
       rdf:about\
=\x22\x22>\x0d\x0a        <d\
c:format>image/s\
vg+xml</dc:forma\
t>\x0d\x0a        <dc:\
type\x0d\x0a          \
 rdf:resource=\x22h\
ttp://purl.org/d\
c/dcmitype/Still\
Image\x22 />\x0d\x0a     \
   <dc:title></d\
c:title>\x0d\x0a      \
</cc:Work>\x0d\x0a    \
</rdf:RDF>\x0d\x0a  </\
metadata>\x0d\x0a  <so\
dipodi:namedview\
\x0d\x0a     inkscape:\
guide-bbox=\x22true\
\x22\x0d\x0a     showguid\
es=\x22true\x22\x0d\x0a     \
inkscape:bbox-no\
des=\x22true\x22\x0d\x0a    \
 inkscape:snap-b\
box=\x22true\x22\x0d\x0a    \
 inkscape:curren\
t-layer=\x22g1687\x22\x0d\
\x0a     inkscape:w\
indow-maximized=\
\x221\x22\x0d\x0a     inksca\
pe:window-y=\x2241\x22\
\x0d\x0a     inkscape:\
window-x=\x22-9\x22\x0d\x0a \
    inkscape:cy=\
\x229.2193472\x22\x0d\x0a   \
  inkscape:cx=\x221\
.7499975\x22\x0d\x0a     \
inkscape:zoom=\x223\
2\x22\x0d\x0a     showgri\
d=\x22false\x22\x0d\x0a     \
id=\x22namedview8\x22\x0d\
\x0a     inkscape:w\
indow-height=\x2210\
01\x22\x0d\x0a     inksca\
pe:window-width=\
\x221920\x22\x0d\x0a     ink\
scape:pageshadow\
=\x222\x22\x0d\x0a     inksc\
ape:pageopacity=\
\x220\x22\x0d\x0a     guidet\
olerance=\x2210\x22\x0d\x0a \
    gridtoleranc\
e=\x2210\x22\x0d\x0a     obj\
ecttolerance=\x2210\
\x22\x0d\x0a     borderop\
acity=\x221\x22\x0d\x0a     \
bordercolor=\x22#66\
6666\x22\x0d\x0a     page\
color=\x22#ffffff\x22 \
/>\x0d\x0a  <defs\x0d\x0a   \
  id=\x22defs3051\x22>\
\x0d\x0a    <inkscape:\
path-effect\x0d\x0a   \
    background_c\
olor=\x22#ffffffff\x22\
\x0d\x0a       backgro\
und=\x22true\x22\x0d\x0a    \
   hide_mask=\x22fa\
lse\x22\x0d\x0a       inv\
ert=\x22false\x22\x0d\x0a   \
    uri=\x22#mask-p\
owermask-path-ef\
fect1654\x22\x0d\x0a     \
  lpeversion=\x221\x22\
\x0d\x0a       is_visi\
ble=\x22true\x22\x0d\x0a    \
   id=\x22path-effe\
ct1654\x22\x0d\x0a       \
effect=\x22powermas\
k\x22 />\x0d\x0a    <styl\
e\x0d\x0a       id=\x22cu\
rrent-color-sche\
me\x22\x0d\x0a       type\
=\x22text/css\x22>\x0d\x0a  \
    .ColorScheme\
-Text {\x0d\x0a       \
 color:#232629;\x0d\
\x0a      }\x0d\x0a      \
</style>\x0d\x0a  </de\
fs>\x0d\x0a  <g\x0d\x0a     \
transform=\x22matri\
x(0.96019972,0,0\
,0.96019972,0.27\
339817,0.3587027\
)\x22\x0d\x0a     id=\x22g16\
87\x22>\x0d\x0a    <path\x0d\
\x0a       style=\x22f\
ill:#232629;fill\
-rule:evenodd;st\
roke:none;stroke\
-width:1px;strok\
e-linecap:butt;s\
troke-linejoin:m\
iter;stroke-opac\
ity:1\x22\x0d\x0a       d\
=\x22M 0.19496303,0\
.10632601 H 3.22\
2636 l -0.0067,5\
.22857129 8.2425\
4,-4e-7 0.0028,-\
5.22843269 2.156\
25,0.00391 2.281\
25,2.28515629 -4\
.31e-4,13.414201\
5 H 13.987456 V \
9.413359 H 2.055\
0291 v 6.396373 \
H 0.19496303 Z\x22\x0d\
\x0a       id=\x22path\
1662\x22\x0d\x0a       so\
dipodi:nodetypes\
=\x22cccccccccccccc\
\x22 />\x0d\x0a    <rect\x0d\
\x0a       style=\x22c\
olor:#000000;ove\
rflow:visible;fi\
ll:#232629;fill-\
opacity:1;stroke\
:none;stroke-wid\
th:2.057;stroke-\
linecap:square;s\
troke-miterlimit\
:4;stroke-dashar\
ray:none;paint-o\
rder:markers str\
oke fill\x22\x0d\x0a     \
  id=\x22rect1667\x22\x0d\
\x0a       width=\x221\
.96875\x22\x0d\x0a       \
height=\x223.46875\x22\
\x0d\x0a       x=\x228.50\
66624\x22\x0d\x0a       y\
=\x220.95913708\x22 />\
\x0d\x0a    <rect\x0d\x0a   \
    style=\x22color\
:#000000;overflo\
w:visible;fill:#\
232629;fill-opac\
ity:1;stroke:non\
e;stroke-width:2\
.02157;stroke-li\
necap:square;str\
oke-miterlimit:4\
;stroke-dasharra\
y:none;paint-ord\
er:markers strok\
e fill\x22\x0d\x0a       \
id=\x22rect1681\x22\x0d\x0a \
      width=\x2215.\
622641\x22\x0d\x0a       \
height=\x221.104854\
3\x22\x0d\x0a       x=\x220.\
19496302\x22\x0d\x0a     \
  y=\x2214.704878\x22 \
/>\x0d\x0a  </g>\x0d\x0a</sv\
g>\x0d\x0a\
"

qt_resource_name = b"\
\x00\x16\
\x06\x1f\xda\xa7\
\x00e\
\x00p\x00o\x00c\x00h\x00-\x00e\x00n\x00c\x00o\x00d\x00e\x00r\x00-\x00s\x00e\x00e\
\x00k\x00.\x00s\x00v\x00g\
\x00\x16\
\x09\x13\xda\x87\
\x00e\
\x00p\x00o\x00c\x00h\x00-\x00e\x00n\x00c\x00o\x00d\x00e\x00r\x00-\x00u\x00n\x00d\
\x00o\x00.\x00s\x00v\x00g\
\x00\x18\
\x0f\xa4\x8b\xc7\
\x00m\
\x00e\x00d\x00i\x00a\x00-\x00p\x00l\x00a\x00y\x00b\x00a\x00c\x00k\x00-\x00s\x00t\
\x00a\x00r\x00t\x00.\x00s\x00v\x00g\
\x00\x17\
\x09\x10g\xc7\
\x00m\
\x00e\x00d\x00i\x00a\x00-\x00p\x00l\x00a\x00y\x00b\x00a\x00c\x00k\x00-\x00s\x00t\
\x00o\x00p\x00.\x00s\x00v\x00g\
\x00\x16\
\x06\x13\xda\x87\
\x00e\
\x00p\x00o\x00c\x00h\x00-\x00e\x00n\x00c\x00o\x00d\x00e\x00r\x00-\x00r\x00e\x00d\
\x00o\x00.\x00s\x00v\x00g\
\x00\x18\
\x0d*I\xc7\
\x00e\
\x00p\x00o\x00c\x00h\x00-\x00e\x00n\x00c\x00o\x00d\x00e\x00r\x00-\x00d\x00e\x00l\
\x00e\x00t\x00e\x00.\x00s\x00v\x00g\
\x00\x1b\
\x0a\x9e4'\
\x00e\
\x00p\x00o\x00c\x00h\x00-\x00e\x00n\x00c\x00o\x00d\x00e\x00r\x00-\x00d\x00u\x00p\
\x00l\x00i\x00c\x00a\x00t\x00e\x00.\x00s\x00v\x00g\
\x00\x17\
\x09\x1f[\x87\
\x00e\
\x00p\x00o\x00c\x00h\x00-\x00e\x00n\x00c\x00o\x00d\x00e\x00r\x00-\x00s\x00p\x00l\
\x00i\x00t\x00.\x00s\x00v\x00g\
\x00\x16\
\x05i\xda\xa7\
\x00e\
\x00p\x00o\x00c\x00h\x00-\x00e\x00n\x00c\x00o\x00d\x00e\x00r\x00-\x00s\x00a\x00v\
\x00e\x00.\x00s\x00v\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x09\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x01\xa6\x00\x00\x00\x00\x00\x01\x00\x00;+\
\x00\x00\x01v\x11\xfb<w\
\x00\x00\x00\xce\x00\x00\x00\x00\x00\x01\x00\x00\x1a\x13\
\x00\x00\x01v\x11\xfb<v\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01v\x11\xfb<\x83\
\x00\x00\x00\x9a\x00\x00\x00\x00\x00\x01\x00\x00\x18\x9a\
\x00\x00\x01k\xb4\xe2\x10\xc2\
\x00\x00\x002\x00\x00\x00\x00\x00\x01\x00\x00\x09u\
\x00\x00\x01v\x11\xfb<\x86\
\x00\x00\x01r\x00\x01\x00\x00\x00\x01\x00\x006\x92\
\x00\x00\x01v\x11\xfb<\x86\
\x00\x00\x016\x00\x01\x00\x00\x00\x01\x00\x001{\
\x00\x00\x01v\x11\xfb<g\
\x00\x00\x01\x00\x00\x00\x00\x00\x00\x01\x00\x00'\xc2\
\x00\x00\x01v\x11\xfb<b\
\x00\x00\x00d\x00\x00\x00\x00\x00\x01\x00\x00\x17&\
\x00\x00\x01k\xb4\xe2\x10\xc3\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()



import numpy as np
import ephyviewer
from  ephyviewer.datasource.epochs import WritableEpochSource


def test_EpochEncoder(interactive=False):
    possible_labels = ['AAA', 'BBB', 'CCC', 'DDD']

    ep_times = np.arange(0, 10., .5)
    ep_durations = np.ones(ep_times.shape) * .25
    ep_labels = np.random.choice(possible_labels, ep_times.size)
    epoch = { 'time':ep_times, 'duration':ep_durations, 'label':ep_labels, 'name': 'MyFactor' }

    source = WritableEpochSource(epoch=epoch, possible_labels=possible_labels)

    app = ephyviewer.mkQApp()
    view = ephyviewer.EpochEncoder(source=source, name='Epoch encoder')

    win = ephyviewer.MainViewer(show_step=False, show_global_xsize=True, debug=False)
    win.add_view(view)

    if interactive:
        win.show()
        app.exec_()
    else:
        # close thread properly
        win.close()


def test_EpochEncoder_settings(interactive=False):
    possible_labels = ['AAA', 'BBB', 'CCC', 'DDD']

    ep_times = np.arange(0, 10., .5)
    ep_durations = np.ones(ep_times.shape) * .25
    ep_labels = np.random.choice(possible_labels, ep_times.size)
    epoch = { 'time':ep_times, 'duration':ep_durations, 'label':ep_labels, 'name': 'MyFactor' }

    source = WritableEpochSource(epoch=epoch, possible_labels=possible_labels)

    app = ephyviewer.mkQApp()
    view = ephyviewer.EpochEncoder(source=source, name='Epoch encoder')

    win = ephyviewer.MainViewer(show_step=True, show_global_xsize=True, debug=False, settings_name='epoch_encode_test1', )
    win.add_view(view)

    if interactive:
        win.show()
        app.exec_()
    else:
        # close thread properly
        win.close()


def test_EpochEncoder_empty(interactive=False):
    possible_labels = ['AAA', 'BBB', 'CCC', 'DDD']

    source = WritableEpochSource(epoch=None, possible_labels=possible_labels)
    source._t_stop = 10 # set to positive value so navigation has a finite range

    app = ephyviewer.mkQApp()
    view = ephyviewer.EpochEncoder(source=source, name='Epoch encoder')

    win = ephyviewer.MainViewer(show_step=False, show_global_xsize=True, debug=False)
    win.add_view(view)

    if interactive:
        win.show()
        app.exec_()
    else:
        # close thread properly
        win.close()


if __name__=='__main__':
    test_EpochEncoder(interactive=True)
    test_EpochEncoder_settings(interactive=True)
    test_EpochEncoder_empty(interactive=True)



import os
import sys


def make_resources_icons(qt_mode):
    with open('icons.qrc','w') as f:
        f.write("""<!DOCTYPE RCC><RCC version="1.0">
    <qresource>
""")
        for p, d, files in os.walk('./'):
            for filename in files:
                if filename.endswith('.png') or filename.endswith('.jpg') or filename.endswith('.svg'):
                    f.write('            <file alias="%s">%s/%s</file>\r\n' % (filename, p[2:],filename) )

        f.write("""    </qresource>
</RCC>
""")

    if qt_mode == 'PySide6':
        os.popen('pyside6-rcc icons.qrc -o icons_PySide6.py')

    elif qt_mode == 'PyQt6':
        # pyrcc6 do not exists!!!!!
        # https://stackoverflow.com/questions/66099225/how-can-resources-be-provided-in-pyqt6-which-has-no-pyrcc
        pass
    elif qt_mode == 'PyQt5':
        os.popen('pyrcc5 icons.qrc -o icons_PyQt5.py')

    elif qt_mode == 'PySide2':
        os.popen('pyside2-rcc icons.qrc -o icons_PySide2.py')

    elif qt_mode == 'PyQt4':
        os.popen('pyrcc4 -py3 icons.qrc -o icons_PyQt4.py')

    else:
        raise ValueError('Cannot build icons for unrecognized Qt bindings: ' + qt_mode)


#------------------------------------------------------------------------------
if __name__ == '__main__' :

    if len(sys.argv) > 1:
        qt_mode = sys.argv[1]
    else:
        from ephyviewer.myqt import QT_MODE as qt_mode

    print('Building icons for Qt bindings:', qt_mode)
    make_resources_icons(qt_mode)



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np

import matplotlib.cm
import matplotlib.colors

from .myqt import QT
import pyqtgraph as pg

from .base import BaseMultiChannelViewer, Base_MultiChannel_ParamController
from .datasource import InMemorySpikeSource, NeoSpikeTrainSource



#make symbol for spikes
from pyqtgraph.graphicsItems.ScatterPlotItem import Symbols
Symbols['|'] = QT.QPainterPath()
Symbols['|'].moveTo(0, -0.5)
Symbols['|'].lineTo(0, 0.5)
Symbols['|'].closeSubpath()




default_params = [
    {'name': 'xsize', 'type': 'float', 'value': 3., 'step': 0.1},
    {'name': 'xratio', 'type': 'float', 'value': 0.3, 'step': 0.1, 'limits': (0,1)},
    {'name': 'scatter_size', 'type': 'float', 'value': 0.8,  'limits': (0,np.inf)},
    {'name': 'background_color', 'type': 'color', 'value': 'k'},
    {'name': 'vline_color', 'type': 'color', 'value': '#FFFFFFAA'},
    {'name': 'label_fill_color', 'type': 'color', 'value': '#222222DD'},
    {'name': 'label_size', 'type': 'int', 'value': 8, 'limits': (1,np.inf)},
    {'name': 'display_labels', 'type': 'bool', 'value': True},
    ]

default_by_channel_params = [
    {'name': 'color', 'type': 'color', 'value': "#55FF00"},
    {'name': 'visible', 'type': 'bool', 'value': True},
    ]




class SpikeTrainViewer_ParamController(Base_MultiChannel_ParamController):
    pass



class DataGrabber(QT.QObject):
    data_ready = QT.pyqtSignal(float, float, object, object)

    def __init__(self, source, parent=None):
        QT.QObject.__init__(self, parent)
        self.source = source

    def on_request_data(self, t_start, t_stop, visibles):
        data = {}
        for e, chan in enumerate(visibles):
            times = self.source.get_chunk_by_time(chan=chan,  t_start=t_start, t_stop=t_stop)
            data[chan] = times
        self.data_ready.emit(t_start, t_stop, visibles, data)


class SpikeTrainViewer(BaseMultiChannelViewer):
    _default_params = default_params
    _default_by_channel_params = default_by_channel_params

    _ControllerClass = SpikeTrainViewer_ParamController

    request_data = QT.pyqtSignal(float, float, object)

    def __init__(self, **kargs):
        BaseMultiChannelViewer.__init__(self, **kargs)

        self.make_params()
        self.set_layout()
        self.make_param_controller()

        self.viewBox.doubleclicked.connect(self.show_params_controller)

        self.initialize_plot()



        self.thread = QT.QThread(parent=self)
        self.datagrabber = DataGrabber(source=self.source)
        self.datagrabber.moveToThread(self.thread)
        self.thread.start()


        self.datagrabber.data_ready.connect(self.on_data_ready)
        self.request_data.connect(self.datagrabber.on_request_data)

        self.params.param('xsize').setLimits((0, np.inf))

    @classmethod
    def from_numpy(cls, all_epochs, name):
        source = InMemorySpikeSource(all_epochs)
        view = cls(source=source, name=name)
        return view

    @classmethod
    def from_neo_spiketrains(cls, neo_spiketrains, name):
        source = NeoSpikeTrainSource(neo_spiketrains)
        view = cls(source=source, name=name)
        return view

    def closeEvent(self, event):
        event.accept()
        self.thread.quit()
        self.thread.wait()


    def initialize_plot(self):
        self.vline = pg.InfiniteLine(angle = 90, movable = False, pen = self.params['vline_color'])
        self.vline.setZValue(1) # ensure vline is above plot elements
        self.plot.addItem(self.vline)

        self.scatter = pg.ScatterPlotItem(size=self.params['scatter_size'], pxMode = False, symbol='|')
        self.plot.addItem(self.scatter)


        self.labels = []
        for c in range(self.source.nb_channel):
            label_name = '{}: {}'.format(c, self.source.get_channel_name(chan=c))
            color = self.by_channel_params.children()[c].param('color').value()
            label = pg.TextItem(label_name, color=color, anchor=(0, 0.5), border=None, fill=self.params['label_fill_color'])
            font = label.textItem.font()
            font.setPointSize(self.params['label_size'])
            label.setFont(font)
            self.plot.addItem(label)
            self.labels.append(label)

        self.viewBox.xsize_zoom.connect(self.params_controller.apply_xsize_zoom)


    def refresh(self):
        xsize = self.params['xsize']
        xratio = self.params['xratio']
        t_start, t_stop = self.t-xsize*xratio , self.t+xsize*(1-xratio)
        visibles, = np.nonzero(self.params_controller.visible_channels)
        self.request_data.emit(t_start, t_stop, visibles)

    def on_data_ready(self, t_start, t_stop, visibles, data):
        self.graphicsview.setBackground(self.params['background_color'])

        self.scatter.clear()
        all_x = []
        all_y = []
        all_brush = []

        for e, chan in enumerate(visibles):
            times = data[chan]

            ypos = visibles.size-e-1

            all_x.append(times)
            all_y.append(np.ones(times.size)*ypos)
            color = self.by_channel_params.children()[chan].param('color').value()
            all_brush.append(np.array([pg.mkPen(color)]*len(times)))


            if self.params['display_labels']:
                self.labels[chan].setPos(t_start, ypos)
                self.labels[chan].show()
                self.labels[chan].setColor(color)

        for chan in range(self.source.nb_channel):
            if not self.params['display_labels'] or chan not in visibles:
                self.labels[chan].hide()

        for label in self.labels:
            label.fill = pg.mkBrush(self.params['label_fill_color'])

        if len(all_x):
            all_x = np.concatenate(all_x)
            all_y = np.concatenate(all_y)
            all_brush = np.concatenate(all_brush)
            self.scatter.setData(x=all_x, y=all_y, pen=all_brush)

        self.vline.setPen(self.params['vline_color'])
        self.vline.setPos(self.t)

        self.plot.setXRange( t_start, t_stop, padding = 0.0)
        self.plot.setYRange(-self.params['scatter_size']/2, self.params['scatter_size']/2 + visibles.size - 1)

    def on_param_change(self, params=None, changes=None):
        for param, change, data in changes:
            if change != 'value': continue
            if param.name()=='scatter_size':
                self.scatter.setSize(self.params['scatter_size'])
            if param.name()=='label_size':
                for label in self.labels:
                    font = label.textItem.font()
                    font.setPointSize(self.params['label_size'])
                    label.setFont(font)

        self.refresh()



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np

import matplotlib.cm
import matplotlib.colors

from .myqt import QT
import pyqtgraph as pg
from pyqtgraph.util.mutex import Mutex

from .base import ViewerBase, BaseMultiChannelViewer,  Base_MultiChannel_ParamController
from .datasource import FrameGrabber, MultiVideoFileSource
from .tools import create_plot_grid

import threading

default_params = [
    {'name': 'nb_column', 'type': 'int', 'value': 4},
    ]

default_by_channel_params = [
    {'name': 'visible', 'type': 'bool', 'value': True},
    ]




#~ class QFrameGrabber(QT.QObject, FrameGrabber):
    #~ frame_ready = QT.pyqtSignal(int, object)
    #~ def on_request_frame(self, video_index, target_frame):
        #~ if self.video_index!=video_index:
            #~ return
        #~ frame = self.get_frame(target_frame)
        #~ if not frame:
            #~ return
        #~ self.frame_ready.emit(self.video_index, frame)


class QFrameGrabber(QT.QObject):
    frame_ready = QT.pyqtSignal(int, object)

    def __init__(self, frame_grabber, video_index, parent=None):
        QT.QObject.__init__(self, parent)

        self.fg = frame_grabber
        self.video_index = video_index

        self.mutex = Mutex()
        self.request_list = []
        self._last_frame = None

    #~ def on_request_frame(self, video_index, target_frame):
        #~ if self.video_index!=video_index:
            #~ return
        #~ frame = self.fg.get_frame(target_frame)
        #~ if frame is None:
            #~ return
        #~ self.frame_ready.emit(self.video_index, frame)

    @property
    def last_frame(self):
        with self.mutex:
            return self._last_frame

    #~ @property
    #~ def active_frame(self):
        #~ return self.fg.active_frame

    #~ @active_frame.setter
    #~ def active_frame(self, value):
        #~ self.fg.active_frame = value

    #~ def queue_request_frame(self, target_frame):
        #~ print('queue_request_frame', threading.get_ident())
        #~ with self.mutex:
            #~ self.request_list.append(target_frame)
            #~ if len(self.request_list)>1:
                #~ self.request_frame.emit()

    def on_request_frame(self, video_index):
        if self.video_index!=video_index:
            return

        #~ print('on_request_frame', threading.get_ident())

        with self.mutex:
            #~ print('len(self.request_list)', len(self.request_list))
            if len(self.request_list)==0:
                return
            target_frame = self.request_list[-1]
            self.request_list = []

            if target_frame == self._last_frame:
                return

        frame = self.fg.get_frame(target_frame)
        with self.mutex:
            self._last_frame = target_frame
            #~ print('new self._last_frame', self._last_frame)
        #~ self.fg.active_frame = target_frame
        self.frame_ready.emit(self.video_index, frame)




class VideoViewer_ParamController(Base_MultiChannel_ParamController):
    pass

#~ class VideoViewer(ViewerBase):
class VideoViewer(BaseMultiChannelViewer):

    _default_params = default_params
    _default_by_channel_params = default_by_channel_params

    _ControllerClass = VideoViewer_ParamController

    request_frame = QT.pyqtSignal(int)

    def __init__(self, **kargs):
        BaseMultiChannelViewer.__init__(self, **kargs)

        self.make_params()
        self.make_param_controller()
        self.set_layout()

        self.qframe_grabbers = []
        self.threads = []
        #~ self.actual_frames = []
        for i, video_filename in enumerate(self.source.video_filenames):
            fg = QFrameGrabber(self.source.frame_grabbers[i], i)
            self.qframe_grabbers.append(fg)
            #~ fg.set_file(video_filename)
            #~ fg.video_index = i
            fg.frame_ready.connect(self.update_frame)

            thread = QT.QThread(parent=self)
            fg.moveToThread(thread)
            thread.start()
            self.threads.append(thread)

            self.request_frame.connect(self.qframe_grabbers[i].on_request_frame)

    @classmethod
    def from_filenames(cls, video_filenames, video_times, name):
        source = MultiVideoFileSource(video_filenames, video_times=video_times)
        view = cls(source=source, name=name)
        return view


    def closeEvent(self, event):
        for i, thread in enumerate(self.threads):
            thread.quit()
            thread.wait()
        event.accept()


    def set_layout(self):
        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)

        self.graphiclayout = pg.GraphicsLayoutWidget()
        self.mainlayout.addWidget(self.graphiclayout)
        self.create_grid()

    def on_param_change(self):
        self.create_grid()
        self.refresh()

    def create_grid(self):
        visible_channels = self.params_controller.visible_channels
        self.plots = create_plot_grid(self.graphiclayout, self.params['nb_column'], visible_channels,
                     ViewBoxClass=pg.ViewBox, vb_params={'lockAspect':True})

        for plot in self.plots:
            plot.showAxis('left', False)
            plot.showAxis('bottom', False)


        self.images = []
        for c in range(self.source.nb_channel):
            if visible_channels[c]:
                image = pg.ImageItem()
                self.plots[c].addItem(image)
                self.images.append(image)
            else:
                self.images.append(None)

    def refresh(self):
        #~ print('videoviewer.refresh', self.t)
        visible_channels = self.params_controller.visible_channels

        #~ print()
        #~ print('refresh t=', self.t)
        for c in range(self.source.nb_channel):
            if visible_channels[c]:
                frame_index = self.source.time_to_frame_index(c, self.t)
                #~ print( 'c', c, 'frame_index', frame_index, 'self.qframe_grabbers[c].last_frame', self.qframe_grabbers[c].last_frame)

                #~ if self.qframe_grabbers[c].active_frame != frame_index:
                #~ print('self.qframe_grabbers[c].last_frame != frame_index', self.qframe_grabbers[c].last_frame != frame_index)
                if self.qframe_grabbers[c].last_frame != frame_index:

                    #~ self.qframe_grabbers[c].active_frame = frame_index
                    #~ self.request_frame.emit(c, frame_index)

                    #~ self.qframe_grabbers[c].queue_request_frame(frame_index)
                    #~ print('enque frame', threading.get_ident(), 'frame_index', frame_index)
                    with self.qframe_grabbers[c].mutex:

                        self.qframe_grabbers[c].request_list.append(frame_index)
                        if len(self.qframe_grabbers[c].request_list)>=1:
                            self.request_frame.emit(c)
                            #~ print('EMIT!!!!!')







    def update_frame(self, video_index, frame):
        #~ print('update_frame', video_index, frame)

        if frame is None:
            self.images[video_index].clear()
        else:
            #TODO : find better solution!!!! to avoid copy
            try:
                # PyAV >= 0.5.3
                img = frame.to_ndarray(format='rgb24')
            except AttributeError:
                # PyAV < 0.5.3
                img = frame.to_nd_array(format='rgb24')
            img = img.swapaxes(0,1)[:,::-1,:]
            #~ print(img.shape, img.dtype)
            self.images[video_index].setImage(img)


        #~ rgba = frame.reformat(frame.width, frame.height, "rgb24", 'itu709')
        #print rgba.to_image().save("test.png")
        # could use the buffer interface here instead, some versions of PyQt don't support it for some reason
        # need to track down which version they added support for it
        #~ bytearray(rgba.planes[0])
        #~ bytesPerPixel  =3
        #~ img = QT.QImage(bytearray(rgba.planes[0]), rgba.width, rgba.height, rgba.width * bytesPerPixel, QT.QImage.Format_RGB888)
        #~ self.images[video_index].setImage(img)

        #img = QtGui.QImage(rgba.planes[0], rgba.width, rgba.height, QtGui.QImage.Format_RGB888)







                #~ if self.actual_frames[i] != new_frame:





                #~ frame = self.source.get_frame(t=t,chan=c)
                #~ self.images[c].setImage(frame)
            #~ else:
                #~ pass



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np
import scipy.fftpack
import scipy.signal

import matplotlib.cm
import matplotlib.colors

from .myqt import QT
import pyqtgraph as pg

from .base import BaseMultiChannelViewer, Base_MultiChannel_ParamController, MyViewBox
from .datasource import InMemoryAnalogSignalSource

from .tools import create_plot_grid


#todo remove this
import time
import threading


default_params = [
    {'name': 'xsize', 'type': 'float', 'value': 3., 'step': 0.1},
    {'name': 'xratio', 'type': 'float', 'value': 0.3, 'step': 0.1, 'limits': (0,1)},
    {'name': 'nb_column', 'type': 'int', 'value': 4},
    {'name': 'background_color', 'type': 'color', 'value': 'k'},
    {'name': 'vline_color', 'type': 'color', 'value': '#FFFFFFAA'},
    {'name': 'colormap', 'type': 'list', 'value': 'viridis', 'values' : ['viridis', 'jet', 'gray', 'hot', ] },
    {'name': 'display_labels', 'type': 'bool', 'value': True},
    {'name': 'show_axis', 'type': 'bool', 'value': True},
    {'name': 'scale_mode', 'type': 'list', 'value': 'same_for_all', 'values' : ['by_channel', 'same_for_all', ] },
    {'name': 'timefreq', 'type': 'group', 'children': [
                    {'name': 'f_start', 'type': 'float', 'value': 3., 'step': 1.},
                    {'name': 'f_stop', 'type': 'float', 'value': 90., 'step': 1.},
                    {'name': 'deltafreq', 'type': 'float', 'value': 3., 'step': 1., 'limits': [0.1, 1.e6]},
                    {'name': 'f0', 'type': 'float', 'value': 2.5, 'step': 0.1},
                    {'name': 'normalisation', 'type': 'float', 'value': 0., 'step': 0.1},]}

    ]

default_by_channel_params = [
    {'name': 'visible', 'type': 'bool', 'value': True},
    {'name': 'clim', 'type': 'float', 'value': .1},
    ]


def generate_wavelet_fourier(len_wavelet, f_start, f_stop, deltafreq, sample_rate, f0, normalisation):
    """
    Compute the wavelet coefficients at all scales and compute its Fourier transform.

    Parameters
    ----------
    len_wavelet : int
        length in samples of the wavelet window
    f_start: float
        First frequency in Hz
    f_stop: float
        Last frequency in Hz
    deltafreq : float
        Frequency interval in Hz
    sample_rate : float
        Sample rate in Hz
    f0 : float
    normalisation : float

    Returns:
    -------

    wf : array
        Fourier transform of the wavelet coefficients (after weighting).
        Axis 0 is time; axis 1 is frequency.
    """
    # compute final map scales
    scales = f0/np.arange(f_start,f_stop,deltafreq)*sample_rate

    # compute wavelet coeffs at all scales
    xi=np.arange(-len_wavelet/2.,len_wavelet/2.)
    xsd = xi[:,np.newaxis] / scales
    wavelet_coefs=np.exp(complex(1j)*2.*np.pi*f0*xsd)*np.exp(-np.power(xsd,2)/2.)

    weighting_function = lambda x: x**(-(1.0+normalisation))
    wavelet_coefs = wavelet_coefs*weighting_function(scales[np.newaxis,:])

    # Transform the wavelet into the Fourier domain
    wf=scipy.fftpack.fft(wavelet_coefs,axis=0)
    wf=wf.conj()

    return wf



class TimeFreqViewer_ParamController(Base_MultiChannel_ParamController):
    some_clim_changed = QT.pyqtSignal()

    def on_channel_visibility_changed(self):
        print('TimeFreqViewer_ParamController.on_channel_visibility_changed')
        self.viewer.create_grid()
        self.viewer.initialize_time_freq()
        self.viewer.refresh()

    def clim_zoom(self, factor):
        #~ print('clim_zoom factor', factor)
        self.viewer.by_channel_params.blockSignals(True)

        for i, p in enumerate(self.viewer.by_channel_params.children()):
            p.param('clim').setValue(p.param('clim').value()*factor)

        self.viewer.by_channel_params.blockSignals(False)
        self.some_clim_changed.emit()

    def compute_auto_clim(self):
        print('compute_auto_clim')
        print(self.visible_channels)

        self.viewer.by_channel_params.blockSignals(True)
        maxs = []
        visibles,  = np.nonzero(self.visible_channels)
        for chan in visibles:
            if chan in self.viewer.last_wt_maps.keys():
                m = np.max(self.viewer.last_wt_maps[chan])
                if self.viewer.params['scale_mode'] == 'by_channel':
                    self.viewer.by_channel_params['ch'+str(chan), 'clim'] = m
                else:
                    maxs.append(m)
        if self.viewer.params['scale_mode'] == 'same_for_all' and len(maxs)>0:
            for chan in visibles:
                self.viewer.by_channel_params['ch'+str(chan), 'clim'] = max(maxs)

        self.viewer.by_channel_params.blockSignals(False)
        self.some_clim_changed.emit()


class TimeFreqWorker(QT.QObject):
    data_ready = QT.pyqtSignal(int, float, float, float,  float, float, object)

    def __init__(self, source,viewer, chan,  parent=None):
        QT.QObject.__init__(self, parent)
        self.source = source
        self.viewer = viewer
        self.chan = chan

    def on_request_data(self, chan, t, t_start, t_stop, visible_channels, worker_params):
        if chan != self.chan:
            return

        if not visible_channels[chan]:
            return

        if self.viewer.t != t:
            print('viewer has moved already', chan, self.viewer.t, t)
            # viewer has moved already
            return

        ds_ratio = worker_params['downsample_ratio']
        sig_chunk_size = worker_params['sig_chunk_size']
        filter_sos = worker_params['filter_sos']
        wavelet_fourrier = worker_params['wavelet_fourrier']
        plot_length = worker_params['plot_length']

        i_start = self.source.time_to_index(t_start)
        #~ print('ds_ratio', ds_ratio)
        #~ print('start', t_start, i_start)

        if ds_ratio>1:
            i_start = i_start - (i_start%ds_ratio)
            #~ print('start', t_start, i_start)

        #clip it
        i_start = max(0, i_start)
        i_start = min(i_start, self.source.get_length())
        if ds_ratio>1:
            #after clip
            i_start = i_start - (i_start%ds_ratio)
        #~ print('start', t_start, i_start)

        i_stop = i_start + sig_chunk_size
        i_stop = min(i_stop, self.source.get_length())


        sigs_chunk = self.source.get_chunk(i_start=i_start, i_stop=i_stop)
        sig = sigs_chunk[:, chan]

        if ds_ratio>1:
            small_sig = scipy.signal.sosfiltfilt(filter_sos, sig)
            small_sig =small_sig[::ds_ratio].copy()  # to ensure continuity
        else:
            small_sig = sig.copy()# to ensure continuity

        left_pad = 0
        if small_sig.shape[0] != wavelet_fourrier.shape[0]:
            #Pad it
            z = np.zeros(wavelet_fourrier.shape[0], dtype=small_sig.dtype)
            left_pad = wavelet_fourrier.shape[0] - small_sig.shape[0]
            z[:small_sig.shape[0]] = small_sig
            small_sig = z


        #avoid border effect
        small_sig -= small_sig.mean()

        #~ print('sig', sig.shape, 'small_sig', small_sig.shape)

        small_sig_f = scipy.fftpack.fft(small_sig)
        if small_sig_f.shape[0] != wavelet_fourrier.shape[0]:
            print('oulala', small_sig_f.shape, wavelet_fourrier.shape)
            #TODO pad with zeros somewhere
            return
        wt_tmp=scipy.fftpack.ifft(small_sig_f[:,np.newaxis]*wavelet_fourrier,axis=0)
        wt = scipy.fftpack.fftshift(wt_tmp,axes=[0])
        wt = np.abs(wt).astype('float32')
        if left_pad>0:
            wt = wt[:-left_pad]
        wt_map = wt[:plot_length]
        #~ wt_map =wt
        #~ print('wt_map', wt_map.shape)


        #~ print('sleep', chan)
        #~ time.sleep(2.)

        #TODO t_start and t_stop wrong
        #~ print('sub_sample_rate', worker_params['sub_sample_rate'])
        #~ print('wanted_size', worker_params['wanted_size'])
        #~ print('plot_length', plot_length)
        #~ print(i_start, i_stop)
        t1 = self.source.index_to_time(i_start)
        t2 = self.source.index_to_time(i_start+wt_map.shape[0]*ds_ratio)
        #~ t2 = self.source.index_to_time(i_stop)
        self.data_ready.emit(chan, t,   t_start, t_stop, t1, t2, wt_map)


class TimeFreqViewer(BaseMultiChannelViewer):

    _default_params = default_params
    _default_by_channel_params = default_by_channel_params

    _ControllerClass = TimeFreqViewer_ParamController

    request_data = QT.pyqtSignal(int, float, float, float, object, object)

    def __init__(self, **kargs):
        BaseMultiChannelViewer.__init__(self, **kargs)

        self.make_params()

        # make all not visible
        self.by_channel_params.blockSignals(True)
        for c in range(self.source.nb_channel):
            self.by_channel_params['ch'+str(c), 'visible'] = c==0
        self.by_channel_params.blockSignals(False)

        self.make_param_controller()
        self.params_controller.some_clim_changed.connect(self.refresh)

        self.set_layout()

        self.change_color_scale()
        self.create_grid()
        self.initialize_time_freq()


        self.last_wt_maps = {}

        self.threads = []
        self.timefreq_makers = []
        for c in range(self.source.nb_channel):
            thread = QT.QThread(parent=self)
            self.threads.append(thread)
            worker = TimeFreqWorker(self.source, self, c)
            self.timefreq_makers.append(worker)
            worker.moveToThread(thread)
            thread.start()


            worker.data_ready.connect(self.on_data_ready)
            self.request_data.connect(worker.on_request_data)

        self.params.param('xsize').setLimits((0, np.inf))

    @classmethod
    def from_numpy(cls, sigs, sample_rate, t_start, name, channel_names=None):
        source = InMemoryAnalogSignalSource(sigs, sample_rate, t_start, channel_names=channel_names)
        view = cls(source=source, name=name)
        return view

    def closeEvent(self, event):
        for i, thread in enumerate(self.threads):
            thread.quit()
            thread.wait()
        event.accept()

    def set_layout(self):
        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)

        self.graphiclayout = pg.GraphicsLayoutWidget()
        self.mainlayout.addWidget(self.graphiclayout)

    def on_param_change(self, params=None, changes=None):
        #~ print('on_param_change')
        #track if new scale mode
        #~ for param, change, data in changes:
            #~ if change != 'value': continue
            #~ if param.name()=='scale_mode':
                #~ self.params_controller.compute_rescale()

        #for simplification everything is recompute
        self.change_color_scale()
        self.create_grid()
        self.initialize_time_freq()
        self.refresh()

    def create_grid(self):
        visible_channels = self.params_controller.visible_channels

        self.plots = create_plot_grid(self.graphiclayout, self.params['nb_column'], visible_channels,
                         ViewBoxClass=MyViewBox,  vb_params={})

        for plot in self.plots:
            if plot is not None:
                plot.vb.doubleclicked.connect(self.show_params_controller)
                plot.vb.ygain_zoom.connect(self.params_controller.clim_zoom)
                # plot.vb.xsize_zoom.connect(self.params_controller.apply_xsize_zoom)

        self.images = []
        self.vlines = []
        for c in range(self.source.nb_channel):
            if visible_channels[c]:
                image = pg.ImageItem()
                self.plots[c].addItem(image)
                self.images.append(image)

                vline = pg.InfiniteLine(angle = 90, movable = False, pen = self.params['vline_color'])
                vline.setZValue(1) # ensure vline is above plot elements
                self.plots[c].addItem(vline)
                self.vlines.append(vline)

            else:
                self.images.append(None)
                self.vlines.append(None)

    def initialize_time_freq(self):
        tfr_params = self.params.param('timefreq')
        sample_rate = self.source.sample_rate

        # we take sample_rate = f_stop*4 or (original sample_rate)
        if tfr_params['f_stop']*4 < sample_rate:
            wanted_sub_sample_rate = tfr_params['f_stop']*4
        else:
            wanted_sub_sample_rate = sample_rate

        # this try to find the best size to get a timefreq of 2**N by changing
        # the sub_sample_rate and the sig_chunk_size
        d = self.worker_params = {}
        d['wanted_size'] = self.params['xsize']
        l = d['len_wavelet'] = int(2**np.ceil(np.log(d['wanted_size']*wanted_sub_sample_rate)/np.log(2)))
        d['sig_chunk_size'] = d['wanted_size']*self.source.sample_rate
        d['downsample_ratio'] = int(np.ceil(d['sig_chunk_size']/l))
        d['sig_chunk_size'] = d['downsample_ratio']*l
        d['sub_sample_rate'] = self.source.sample_rate/d['downsample_ratio']
        d['plot_length'] = int(d['wanted_size']*d['sub_sample_rate'])

        d['wavelet_fourrier'] = generate_wavelet_fourier(d['len_wavelet'], tfr_params['f_start'], tfr_params['f_stop'],
                            tfr_params['deltafreq'], d['sub_sample_rate'], tfr_params['f0'], tfr_params['normalisation'])

        if d['downsample_ratio']>1:
            n = 8
            q = d['downsample_ratio']
            d['filter_sos'] = scipy.signal.cheby1(n, 0.05, 0.8 / q, output='sos')
        else:
            d['filter_sos'] = None

    def change_color_scale(self):
        N = 512
        cmap_name = self.params['colormap']
        cmap = matplotlib.cm.get_cmap(cmap_name , N)

        lut = []
        for i in range(N):
            r,g,b,_ =  matplotlib.colors.ColorConverter().to_rgba(cmap(i))
            lut.append([r*255,g*255,b*255])
        self.lut = np.array(lut, dtype='uint8')

    def auto_scale(self):
        print('auto_scale', self.params['scale_mode'])
        self.params_controller.compute_auto_clim()
        self.refresh()

    def refresh(self):
        #~ print('TimeFreqViewer.refresh', self.t)
        visible_channels = self.params_controller.visible_channels

        xsize = self.params['xsize']
        xratio = self.params['xratio']
        t_start, t_stop = self.t-xsize*xratio , self.t+xsize*(1-xratio)

        for c in range(self.source.nb_channel):
            if visible_channels[c]:
                self.request_data.emit(c, self.t, t_start, t_stop, visible_channels, self.worker_params)

        self.graphiclayout.setBackground(self.params['background_color'])

    def on_data_ready(self, chan, t,   t_start, t_stop, t1,t2, wt_map):
        if not self.params_controller.visible_channels[chan]:
            return

        if self.images[chan] is None:
            return

        self.last_wt_maps[chan] = wt_map
        f_start = self.params['timefreq', 'f_start']
        f_stop = self.params['timefreq', 'f_stop']

        image = self.images[chan]
        #~ print(t_start, f_start,self.worker_params['wanted_size'], f_stop-f_start)

        #~ image.updateImage(wt_map)
        clim = self.by_channel_params['ch{}'.format(chan), 'clim']
        image.setImage(wt_map, lut=self.lut, levels=[0, clim])
        image.setRect(QT.QRectF(t1, f_start,t2-t1, f_stop-f_start))

        #TODO
        # display_labels

        self.vlines[chan].setPos(t)
        self.vlines[chan].setPen(self.params['vline_color'])
        plot = self.plots[chan]
        plot.setXRange(t_start, t_stop, padding = 0.0)
        plot.setYRange(f_start, f_stop, padding = 0.0)

        if self.params['display_labels']:
            ch_name = '{}: {}'.format(chan, self.source.get_channel_name(chan=chan))
            self.plots[chan].setTitle(ch_name)
        else:
            self.plots[chan].setTitle(None)


        self.plots[chan].showAxis('left', self.params['show_axis'])
        self.plots[chan].showAxis('bottom', self.params['show_axis'])



import numpy as np
import ephyviewer
from  ephyviewer.tests.testing_tools import make_fake_signals, make_fake_signals_with_scatter

import neo
import quantities as pq


def test_traceviewer(interactive=False):
    source = make_fake_signals()


    app = ephyviewer.mkQApp()
    view = ephyviewer.TraceViewer(source=source, name='trace')
    #~ view.refresh()


    win = ephyviewer.MainViewer(debug=True, show_auto_scale=True)
    win.add_view(view)

    if interactive:
        win.show()
        app.exec_()
    else:
        # close thread properly
        win.close()


def test_traceviewer_with_scatter(interactive=False):
    source = make_fake_signals_with_scatter()

    #~ exit()

    app = ephyviewer.mkQApp()
    view = ephyviewer.TraceViewer(source=source, name='trace_with_scatter')
    #~ view.refresh()


    win = ephyviewer.MainViewer(debug=True, show_auto_scale=True)
    win.add_view(view)

    if interactive:
        win.show()
        app.exec_()
    else:
        # close thread properly
        win.close()



def test_traceviewer_cls_method_numpy(interactive=False):
    sigs = np.random.rand(100000,16)
    sample_rate = 1000.
    t_start = 0.

    app = ephyviewer.mkQApp()
    view = ephyviewer.TraceViewer.from_numpy(sigs, sample_rate, t_start, 'sigs')

    win = ephyviewer.MainViewer(debug=True, show_auto_scale=True)
    win.add_view(view)

    if interactive:
        win.show()
        app.exec_()
    else:
        # close thread properly
        win.close()

def test_traceviewer_cls_method_neo(interactive=False):
    sigs = np.random.rand(100000,16)
    sample_rate = 1000.
    t_start = 0.

    neo_anasig = neo.AnalogSignal(sigs*pq.mV, sampling_rate=sample_rate*pq.Hz, t_start=0*pq.s, copy=True)
    print(neo_anasig)


    app = ephyviewer.mkQApp()

    view = ephyviewer.TraceViewer.from_neo_analogsignal(neo_anasig, 'sigs')
    win = ephyviewer.MainViewer(debug=True, show_auto_scale=True)
    win.add_view(view)

    if interactive:
        win.show()
        app.exec_()
    else:
        # close thread properly
        win.close()



if __name__=='__main__':
    test_traceviewer(interactive=True)
    test_traceviewer_with_scatter(interactive=True)
    test_traceviewer_cls_method_numpy(interactive=True)
    test_traceviewer_cls_method_neo(interactive=True)




import ephyviewer
import numpy as np
import os

# get_tdt_test_files
from ephyviewer.tests.testing_tools import get_blackrock_files


def test_neoviewer(interactive=False):
    #~ from neo.rawio.tdtrawio import TdtRawIO

    #~ local_tdt_folder = get_tdt_test_files()
    #~ neorawio = TdtRawIO(dirname=local_tdt_folder)
    #~ neorawio.parse_header()
    #~ print(neorawio)

    from neo.rawio.blackrockrawio import BlackrockRawIO
    filename = get_blackrock_files()
    neorawio = BlackrockRawIO(filename=filename)
    neorawio.parse_header()
    print(neorawio)
    

    sources = ephyviewer.get_sources_from_neo_rawio(neorawio)
    print(sources)


    app = ephyviewer.mkQApp()
    win = ephyviewer.MainViewer(debug=True, show_auto_scale=True)

    for i, sig_source in enumerate(sources['signal']):
        view = ephyviewer.TraceViewer(source=sig_source, name='signal {}'.format(i))
        win.add_view(view)

    for i, ep_source in enumerate(sources['epoch']):
        view = ephyviewer.EpochViewer(source=ep_source, name='epochs')
        win.add_view(view)

    if interactive:
        win.show()
        app.exec_()
    else:
        # close thread properly
        win.close()



if __name__=='__main__':
    test_neoviewer(interactive=True)



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)

import gc
from collections import OrderedDict

from .myqt import QT, QT_MODE
import pyqtgraph as pg


from .mainviewer import MainViewer, compose_mainviewer_from_sources
#~ from .traceviewer import TraceViewer
#~ from .epochviewer import EpochViewer
#~ from .eventlist import EventList
#~ from .spiketrainviewer import SpikeTrainViewer
from.tools import get_dict_from_group_param


from .datasource import get_sources_from_neo_rawio

# browse neo.rawios and add some gui_params
from neo.rawio import rawiolist
rawio_gui_params = {}
rawio_gui_params['RawBinarySignal'] = [
        {'name': 'dtype', 'type': 'list', 'values':['int16', 'uint16', 'float32', 'float64']},
        {'name': 'nb_channel', 'type': 'int', 'value':1},
        {'name': 'sampling_rate', 'type': 'float', 'value':10000., 'step': 1000., 'suffix': 'Hz', 'siPrefix': True},
        {'name': 'bytesoffset', 'type': 'int', 'value':0},
    ]

all_neo_rawio_dict = OrderedDict()
for rawio_class in rawiolist:
    name = rawio_class.__name__.replace('RawIO', '')
    #~ print(name)
    rawio_class.name = name
    rawio_class.gui_params = rawio_gui_params.get(name, None)
    assert name not in all_neo_rawio_dict
    all_neo_rawio_dict[name] = rawio_class




class WindowManager():
    def __init__(self):

        self.windows = []

    def open_dialog(self):
        dia = RawNeoOpenDialog()
        if dia.exec_():
            self.load_dataset(**dia.final_params)

    def load_dataset(self, neo_rawio_class=None, file_or_dir_names=[], io_params={}):
        #~ assert len(file_or_dir_names)==1
        if len(file_or_dir_names)!=1:
            return

        kargs = dict()
        if neo_rawio_class.rawmode.endswith('-file'):
            kargs['filename'] = file_or_dir_names[0]
        elif neo_rawio_class.rawmode.endswith('-dir'):
            kargs['dirname'] = file_or_dir_names[0]
        kargs.update(io_params)

        neorawio = neo_rawio_class(**kargs)
        neorawio.parse_header()
        print(neorawio)

        # initialize a new main window and populate viewers
        navigation_params = {}
        win = MainViewer(debug=False, settings_name=None, **navigation_params)
        sources = get_sources_from_neo_rawio(neorawio)
        compose_mainviewer_from_sources(sources, mainviewer=win)

        # add menus
        file_menu = win.menuBar().addMenu(win.tr("File"))
        do_open = QT.QAction('&Open', win, shortcut = "Ctrl+O")
        do_open.triggered.connect(self.open_dialog)
        file_menu.addAction(do_open)

        # store a reference to the window to prevent it from going out of scope
        # immediately and being destroyed
        self.windows.append(win)

        # delete window on close so that memory and file resources are released
        win.setAttribute(QT.WA_DeleteOnClose, True)
        win.destroyed.connect(
            lambda *args, i=len(self.windows)-1: self.free_resources(i))

        win.show()

        #~ for i, sig_source in enumerate(sources['signal']):
            #~ view = TraceViewer(source=sig_source, name='signal {}'.format(i))
            #~ view.params['scale_mode'] = 'same_for_all'
            #~ view.params['display_labels'] = True
            #~ view.auto_scale()
            #~ if i==0:
                #~ self.add_view(view)
            #~ else:
                #~ self.add_view(view, tabify_with='signal {}'.format(i-1))


        #~ for i, spike_source in enumerate(sources['spike']):
            #~ view = SpikeTrainViewer(source=spike_source, name='spikes')
            #~ self.add_view(view)

        #~ for i, ep_source in enumerate(sources['epoch']):
            #~ view = EpochViewer(source=ep_source, name='epochs')
            #~ self.add_view(view)

            #~ view = EventList(source=ep_source, name='Event list')
            #~ self.add_view(view, location='bottom',  orientation='horizontal')

    def free_resources(self, i):
        """
        Run garbage collection to unlock files and free memory for the closed
        window with index ``i``.
        Data files opened by Neo in lazy mode remain locked for as long as the
        RawIO objects pointing to them exist in memory. Normally such objects
        would be automatically garbage collected when they go out of scope,
        i.e., when the window that created them is closed. However, due to an
        issue in Neo, circular references to these objects are always created,
        so they persist even after the window is closed. This function performs
        a manual garbage collection after a window has been closed to clean up
        any lingering Neo objects that keep files locked. For more info about
        the issue, see https://github.com/NeuralEnsemble/python-neo/issues/684.
        """

        # first remove the last remaining reference to the closed window
        self.windows[i] = None

        # run garbage collection
        gc.collect()




class RawNeoOpenDialog(QT.QDialog):
    def __init__(self, parent=None):
        QT.QDialog.__init__(self, parent = parent)

        self.setWindowTitle('Open file or dir with neo.rawio')
        self.setModal(True)

        self.resize(600, 300)

        layout = QT.QVBoxLayout()
        self.setLayout(layout)


        #for step 'source_type'
        self.tree_params = pg.parametertree.ParameterTree(parent=self)
        self.tree_params.header().hide()
        layout.addWidget(self.tree_params)
        self.tree_params.hide()

        # for filenames
        self.but_addfiles = QT.QPushButton('Select file')
        layout.addWidget(self.but_addfiles)
        self.but_addfiles.clicked.connect(self.on_addfiles)
        self.but_addfiles.hide()
        self.list_files = QT.QListWidget()
        layout.addWidget(self.list_files)
        self.list_files.hide()
        layout.addStretch()

        self.but_next = QT.QPushButton('Next >')
        layout.addWidget(self.but_next)
        self.but_next.clicked.connect(self.validate_step)

        self.steps = ['source_type', 'filenames', 'io_params', ]
        self.display_step(self.steps[0])

        self.final_params = {}

    def display_step(self, step):
        self.actual_step = step

        if step=='source_type':
            source_types = list(all_neo_rawio_dict.keys())
            params = [{'name': 'format', 'type': 'list', 'values':source_types},
                            ]
            self.step_params = pg.parametertree.Parameter.create(name='Select format', type='group', children = params)
            self.tree_params.setParameters(self.step_params, showTop=True)
            self.tree_params.show()

        elif step=='filenames':
            self.but_addfiles.show()
            self.list_files.show()

        elif step=='io_params':
            params = self.neo_rawio_class.gui_params
            self.step_params = pg.parametertree.Parameter.create(name='Set params', type='group', children=params)
            self.tree_params.setParameters(self.step_params, showTop=True)
            self.tree_params.show()


    def validate_step(self):
        #~ print('validate_step', self.actual_step)
        step = self.actual_step
        if step=='source_type':
            self.neo_rawio_class = all_neo_rawio_dict[self.step_params['format']]
            self.final_params['neo_rawio_class'] = self.neo_rawio_class
            self.tree_params.hide()
            self.display_step('filenames')

        elif step=='filenames':
            file_or_dir_names = [self.list_files.item(i).text() for i in range(self.list_files.count())]
            if len(file_or_dir_names)==0: return
            self.final_params['file_or_dir_names'] = file_or_dir_names
            self.but_addfiles.hide()
            self.list_files.hide()

            if self.neo_rawio_class.gui_params is None:
                self.final_params['io_params'] = {}
                self.accept()

            else:
                self.display_step('io_params')

        elif step=='io_params':
            self.final_params['io_params'] = get_dict_from_group_param(self.step_params)
            self.tree_params.hide()
            self.display_step('mk_chan_grp')
            self.accept()


    def on_addfiles(self):
        #~ print('on_addfiles')
        #~ print(self.neo_rawio_class)
        #~ print(self.neo_rawio_class.rawmode)

        if self.neo_rawio_class.rawmode.endswith('-file'):
            fd = QT.QFileDialog(fileMode=QT.QFileDialog.ExistingFiles, acceptMode=QT.QFileDialog.AcceptOpen)
            #Todo play with rawio.extentsions
            fd.setNameFilters(['All (*)'])
        elif self.neo_rawio_class.rawmode.endswith('-dir'):
            fd = QT.QFileDialog(fileMode=QT.QFileDialog.DirectoryOnly, acceptMode=QT.QFileDialog.AcceptOpen)
            #~ fd.setNameFilters(['All (*)'])

        fd.setViewMode(QT.QFileDialog.Detail)
        if fd.exec_():
            filenames = fd.selectedFiles()
            self.list_files.addItems(filenames)



import numpy as np
import scipy.fftpack
import scipy.signal

import matplotlib.cm
import matplotlib.colors

from .myqt import QT
import pyqtgraph as pg

from .base import BaseMultiChannelViewer, Base_MultiChannel_ParamController, MyViewBox
from .datasource import InMemoryAnalogSignalSource

from .tools import create_plot_grid, get_dict_from_group_param


#todo remove this
import time
import threading



default_params = [
    {'name': 'xsize', 'type': 'float', 'value': 10., 'step': 0.1},
    {'name': 'xratio', 'type': 'float', 'value': 0.3, 'step': 0.1, 'limits': (0,1)},
    {'name': 'nb_column', 'type': 'int', 'value': 4},
    {'name': 'background_color', 'type': 'color', 'value': 'k'},
    {'name': 'vline_color', 'type': 'color', 'value': '#FFFFFFAA'},
    {'name': 'colormap', 'type': 'list', 'value': 'viridis', 'values' : ['inferno', 'viridis', 'jet', 'gray', 'hot', ] },
    {'name': 'scale_mode', 'type': 'list', 'value': 'same_for_all', 'values' : ['by_channel', 'same_for_all', ] },
    {'name': 'display_labels', 'type': 'bool', 'value': True},
    {'name': 'show_axis', 'type': 'bool', 'value': True},
    {'name': 'scalogram', 'type': 'group', 'children': [
                    {'name': 'binsize', 'type': 'float', 'value': 0.01, 'step': .01, 'limits': (0,np.inf)},
                    {'name': 'overlapratio', 'type': 'float', 'value': 0., 'step': .05, 'limits': (0., 0.95)},
                    {'name': 'scaling', 'type': 'list', 'value': 'density', 'values' : ['density', 'spectrum'] },
                    {'name': 'mode', 'type': 'list', 'value': 'psd', 'values' : ['psd'] },
                    {'name': 'detrend', 'type': 'list', 'value': 'constant', 'values' : ['constant'] },
                    {'name': 'scale', 'type': 'list', 'value': 'dB', 'values' : ['dB', 'linear'] },
                    

                ]
            }
                    
    ]

default_by_channel_params = [
    {'name': 'visible', 'type': 'bool', 'value': True},
    {'name': 'clim_min', 'type': 'float', 'value': -10},
    {'name': 'clim_max', 'type': 'float', 'value': 0.},
    ]
    
    
    
class SpectrogramViewer_ParamController(Base_MultiChannel_ParamController):
    some_clim_changed = QT.pyqtSignal()

    def on_channel_visibility_changed(self):
        #~ print('SpectrogramViewer_ParamController.on_channel_visibility_changed')
        self.viewer.create_grid()
        self.viewer.refresh()

    def clim_zoom(self, factor):
        print('clim_zoom factor', factor)
        self.viewer.by_channel_params.blockSignals(True)

        for i, p in enumerate(self.viewer.by_channel_params.children()):
            # p.param('clim').setValue(p.param('clim').value()*factor)
            min_ = p['clim_min']
            max_ = p['clim_max']
            d = max_ - min_
            m = (min_ + max_) / 2.
            p['clim_min'] = m - d/2. * factor
            p['clim_max'] = m + d/2. * factor

        self.viewer.by_channel_params.blockSignals(False)
        self.some_clim_changed.emit()

    def compute_auto_clim(self):
        #~ print('compute_auto_clim')
        #~ print(self.visible_channels)

        self.viewer.by_channel_params.blockSignals(True)
        mins = []
        maxs = []
        visibles,  = np.nonzero(self.visible_channels)
        for chan in visibles:
            if chan in self.viewer.last_Sxx.keys():
                min_ = np.min(self.viewer.last_Sxx[chan])
                max_ = np.max(self.viewer.last_Sxx[chan])
                if self.viewer.params['scale_mode'] == 'by_channel':
                    self.viewer.by_channel_params['ch'+str(chan), 'clim_min'] = min_
                    self.viewer.by_channel_params['ch'+str(chan), 'clim_max'] = max_
                    
                else:
                    mins.append(min_)
                    maxs.append(max_)
                    
        if self.viewer.params['scale_mode'] == 'same_for_all' and len(maxs)>0:
            for chan in visibles:
                self.viewer.by_channel_params['ch'+str(chan), 'clim_min'] = np.min(mins)
                self.viewer.by_channel_params['ch'+str(chan), 'clim_max'] = np.max(maxs)

        self.viewer.by_channel_params.blockSignals(False)
        self.some_clim_changed.emit()



class SpectrogramWorker(QT.QObject):
    data_ready = QT.pyqtSignal(int, float, float, float,  float, float, object)

    def __init__(self, source,viewer, chan,  parent=None):
        QT.QObject.__init__(self, parent)
        self.source = source
        self.viewer = viewer
        self.chan = chan

    def on_request_data(self, chan, t, t_start, t_stop, visible_channels, worker_params):
        if chan != self.chan:
            return

        if not visible_channels[chan]:
            return

        if self.viewer.t != t:
            print('viewer has moved already', chan, self.viewer.t, t)
            # viewer has moved already
            return

        binsize = worker_params['binsize']
        overlapratio = worker_params['overlapratio']
        scaling = worker_params['scaling']
        detrend = worker_params['detrend']
        mode = worker_params['mode']
        

        i_start = self.source.time_to_index(t_start)
        i_stop = self.source.time_to_index(t_stop)
        # clip
        i_start = min(max(0, i_start), self.source.get_length())
        i_stop = min(max(0, i_stop), self.source.get_length())
        

        sr = self.source.sample_rate
        nperseg = int(binsize * sr)
        noverlap = int(overlapratio * nperseg)
        
        if noverlap >= nperseg:
            noverlap = noverlap - 1
        print(nperseg, noverlap)

        if nperseg== 0 or (i_stop - i_start) < nperseg:
            # too short
            t1, t2 = t_start, t_stop
            Sxx = None
            self.data_ready.emit(chan, t,   t_start, t_stop, t1, t2, Sxx)
        else:

            sigs_chunk = self.source.get_chunk(i_start=i_start, i_stop=i_stop)
            sig = sigs_chunk[:, chan]

            freqs, times, Sxx = scipy.signal.spectrogram(sig, fs=sr,nperseg=nperseg, noverlap=noverlap,
                        detrend=detrend, scaling=scaling, mode=mode)
            
            if  worker_params['scale'] == 'dB':
                if mode == 'psd':
                    Sxx = 10. * np.log10(Sxx)
            
            if len(times) >=2:
                bin_slide = times[1] - times[0]
                t1 = self.source.index_to_time(i_start) + times[0] - bin_slide /2.
                t2 = self.source.index_to_time(i_start) + times[-1] + bin_slide /2.
                self.data_ready.emit(chan, t,   t_start, t_stop, t1, t2, Sxx)
            else:
                t1, t2 = t_start, t_stop
                Sxx = None
                self.data_ready.emit(chan, t,   t_start, t_stop, t1, t2, Sxx)





class SpectrogramViewer(BaseMultiChannelViewer):

    _default_params = default_params
    _default_by_channel_params = default_by_channel_params

    _ControllerClass = SpectrogramViewer_ParamController

    request_data = QT.pyqtSignal(int, float, float, float, object, object)

    def __init__(self, **kargs):
        BaseMultiChannelViewer.__init__(self, **kargs)

        self.make_params()

        # make all not visible
        self.by_channel_params.blockSignals(True)
        for c in range(self.source.nb_channel):
            self.by_channel_params['ch'+str(c), 'visible'] = c==0
        self.by_channel_params.blockSignals(False)

        self.make_param_controller()
        self.params_controller.some_clim_changed.connect(self.refresh)

        self.set_layout()

        self.change_color_scale()
        self.create_grid()


        self.last_Sxx = {}

        self.threads = []
        self.timefreq_makers = []
        for c in range(self.source.nb_channel):
            thread = QT.QThread(parent=self)
            self.threads.append(thread)
            worker = SpectrogramWorker(self.source, self, c)
            self.timefreq_makers.append(worker)
            worker.moveToThread(thread)
            thread.start()
            
            self.last_Sxx[c] = None


            worker.data_ready.connect(self.on_data_ready)
            self.request_data.connect(worker.on_request_data)

        self.params.param('xsize').setLimits((0, np.inf))

    @classmethod
    def from_numpy(cls, sigs, sample_rate, t_start, name, channel_names=None):
        source = InMemoryAnalogSignalSource(sigs, sample_rate, t_start, channel_names=channel_names)
        view = cls(source=source, name=name)
        return view

    def closeEvent(self, event):
        for i, thread in enumerate(self.threads):
            thread.quit()
            thread.wait()
        event.accept()

    def set_layout(self):
        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)

        self.graphiclayout = pg.GraphicsLayoutWidget()
        self.mainlayout.addWidget(self.graphiclayout)

    def on_param_change(self, params=None, changes=None):
        #~ print('on_param_change')
        #track if new scale mode
        #~ for param, change, data in changes:
            #~ if change != 'value': continue
            #~ if param.name()=='scale_mode':
                #~ self.params_controller.compute_rescale()

        #for simplification everything is recompute
        self.change_color_scale()
        self.create_grid()
        self.refresh()

    def create_grid(self):
        visible_channels = self.params_controller.visible_channels

        self.plots = create_plot_grid(self.graphiclayout, self.params['nb_column'], visible_channels,
                         ViewBoxClass=MyViewBox,  vb_params={})

        for plot in self.plots:
            if plot is not None:
                plot.vb.doubleclicked.connect(self.show_params_controller)
                plot.vb.ygain_zoom.connect(self.params_controller.clim_zoom)
                # plot.vb.xsize_zoom.connect(self.params_controller.apply_xsize_zoom)

        self.images = []
        self.vlines = []
        for c in range(self.source.nb_channel):
            if visible_channels[c]:
                image = pg.ImageItem()
                self.plots[c].addItem(image)
                self.images.append(image)

                vline = pg.InfiniteLine(angle = 90, movable = False, pen = self.params['vline_color'])
                vline.setZValue(1) # ensure vline is above plot elements
                self.plots[c].addItem(vline)
                self.vlines.append(vline)

            else:
                self.images.append(None)
                self.vlines.append(None)

    def change_color_scale(self):
        N = 512
        cmap_name = self.params['colormap']
        cmap = matplotlib.cm.get_cmap(cmap_name , N)

        lut = []
        for i in range(N):
            r,g,b,_ =  matplotlib.colors.ColorConverter().to_rgba(cmap(i))
            lut.append([r*255,g*255,b*255])
        self.lut = np.array(lut, dtype='uint8')

    def auto_scale(self):
        #~ print('auto_scale', self.params['scale_mode'])
        self.params_controller.compute_auto_clim()
        self.refresh()

    def refresh(self):
        #~ print('TimeFreqViewer.refresh', self.t)
        visible_channels = self.params_controller.visible_channels

        xsize = self.params['xsize']
        xratio = self.params['xratio']
        t_start, t_stop = self.t-xsize*xratio , self.t+xsize*(1-xratio)
        
        worker_params = get_dict_from_group_param(self.params.param('scalogram'))
        #~ print('worker_params', worker_params)
        
        
        for c in range(self.source.nb_channel):
            if visible_channels[c]:
                self.request_data.emit(c, self.t, t_start, t_stop, visible_channels, worker_params)

        self.graphiclayout.setBackground(self.params['background_color'])

    def on_data_ready(self, chan, t,   t_start, t_stop, t1,t2, Sxx):
        if not self.params_controller.visible_channels[chan]:
            return

        if self.images[chan] is None:
            return
        
        
        self.last_Sxx[chan] = Sxx
        
        image = self.images[chan]
        
        if Sxx is None:
            image.hide()
            return
        
        image.show()
        
        f_start = 0
        f_stop = self.source.sample_rate / 2.
        #~ f_start = self.params['timefreq', 'f_start']
        #~ f_stop = self.params['timefreq', 'f_stop']

        
        #~ print(t_start, f_start,self.worker_params['wanted_size'], f_stop-f_start)

        #~ image.updateImage(wt_map)
        clim_min = self.by_channel_params['ch{}'.format(chan), 'clim_min']
        clim_max = self.by_channel_params['ch{}'.format(chan), 'clim_max']
        
        #~ clim = np.max(Sxx)
        image.setImage(Sxx.T, lut=self.lut, levels=[clim_min, clim_max])
        #~ image.setImage(Sxx.T, lut=self.lut, levels=[ np.min(Sxx), np.max(Sxx)])
        image.setRect(QT.QRectF(t1, f_start,t2-t1, f_stop-f_start))

        #TODO
        # display_labels

        self.vlines[chan].setPos(t)
        self.vlines[chan].setPen(self.params['vline_color'])
        plot = self.plots[chan]
        plot.setXRange(t_start, t_stop, padding = 0.0)
        plot.setYRange(f_start, f_stop, padding = 0.0)

        if self.params['display_labels']:
            ch_name = '{}: {}'.format(chan, self.source.get_channel_name(chan=chan))
            self.plots[chan].setTitle(ch_name)
        else:
            self.plots[chan].setTitle(None)


        self.plots[chan].showAxis('left', self.params['show_axis'])
        self.plots[chan].showAxis('bottom', self.params['show_axis'])



